__all__ = [
    "ItemContainer",
]
import re
from typing import Union
import warnings

import numpy as np
import numpy.typing as npt

from riptable.rt_enum import ColumnAttribute

ATTRIBUTE_LABEL = "Label"
ATTRIBUTE_SUMMARY = "Right"
ATTRIBUTE_FOOTER = "Footer"
ATTRIBUTE_MARGIN_COLUMN = "MarginColumn"
ATTRIBUTE_NUMBER_OF_FOOTER_ROWS = "NumberOfFooterRows"


class ItemAttribute:
    """
    An attribute about an item which, in turn, contains attributes in the
    form of Python attributes, set and retrieved using setattr() and getattr()
    """

    ATTRIB_EXCLUSION_LIST = "copy"

    def __repr__(self, indent: int = 2):
        result = self.__class__.__name__ + "\n"
        for k, v in self._attribs():
            result += " " * indent + k + ": " + str(v)
        result += "\n"
        return result

    def _attribs(self):
        """
        Returns all attributes dynamically set for this ItemAttribute..

        NOTE: Add to the ATTRIB_EXCLUSION_LIST all method or property names statically
        added to ItemAttribute that don't begin with '_'.
        :return:
        """
        return [
            (k, getattr(self, k))
            for k in dir(self)
            if (not k.startswith("_") and k not in ItemAttribute.ATTRIB_EXCLUSION_LIST)
        ]

    def copy(self):
        """
        Performs a deep copy of the ItemAttribute, including all values
        of any dynamically added attributes.
        :return:
        """
        attrib = ItemAttribute()
        for k, v in self._attribs():
            setattr(attrib, k, v.copy() if hasattr(v, "copy") else v)
        return attrib


class ItemContainer:
    "Container for items in Struct -- all values are tuples with an attribute"

    def __init__(self, *args, **kwds):
        """Initialize an IC"""
        self._items = {}
        self._items.update(*args, **kwds)

    def __getitem__(self, key):
        return self._items[key]

    def __setitem__(self, key, value):
        "ic.__setitem__(i, y) <==> ic[i]=y"
        self._items[key] = value

    def __delitem__(self, key):
        "ic.__delitem__(y) <==> del ic[y]"
        # Deleting an existing item uses self.__map to find the link which is
        # then removed by updating the links in the predecessor and successor nodes.
        del self._items[key]

    # def __iter__(self):
    #    'od.__iter__() <==> iter(od)'
    #    # Traverse the linked list in order.
    #    root = self.__root
    #    curr = root[NEXT]
    #    while curr is not root:
    #        yield curr[KEY]
    #        curr = curr[NEXT]

    # def __reversed__(self):
    #    'od.__reversed__() <==> reversed(od)'
    #    # Traverse the linked list in reverse order.
    #    root = self.__root
    #    curr = root[PREV]
    #    while curr is not root:
    #        yield curr[KEY]
    #        curr = curr[PREV]

    # def __reduce__(self):
    #    'Return state information for pickling'
    #    items = [[k, self[k]] for k in self]
    #    tmp = self.__map, self.__root
    #    del self.__map, self.__root
    #    inst_dict = vars(self).copy()
    #    self.__map, self.__root = tmp
    #    if inst_dict:
    #        return (self.__class__, (items,), inst_dict)
    #    return self.__class__, (items,)

    def clear(self):
        self._items.clear()

    def __contains__(self, *args):
        return self._items.__contains__(*args)

    def __next__(self):
        return self._items.__next__()

    def __len__(self):
        return self._items.__len__()

    def __iter__(self):
        # return self._items.__iter__()
        return iter(self._items)

    def items(self):
        return self._items.items()

    def values(self):
        return self._items.values()

    def keys(self):
        # how to best do this?
        return list(self._items.keys())

    def setdefault(self, *args):
        return self._items.setdefault(*args)

    def update(self, *args):
        return self._items.update(*args)

    def pop(self, *args):
        return self._items.pop(*args)

    # setdefault = MutableMapping.setdefault
    # update = MutableMapping.update
    # pop = MutableMapping.pop
    # keys = MutableMapping.keys
    # values = MutableMapping.values
    # items = MutableMapping.items
    # __ne__ = MutableMapping.__ne__

    # def popitem(self, last=True):
    #    '''od.popitem() -> (k, v), return and remove a (key, value) pair.
    #    Pairs are returned in LIFO order if last is true or FIFO order if false.

    #    '''
    #    if not self:
    #        raise KeyError('dictionary is empty')
    #    key = next(reversed(self) if last else iter(self))
    #    value = self.pop(key)
    #    return key, value

    # -----------------------------------------
    def __repr__(self):
        "ic.__repr__() <==> repr(ic)"
        if not self:
            return "%s()" % (self.__class__.__name__,)
        return "%s(%r)" % (self.__class__.__name__, self._items.items())

    # -----------------------------------------
    def copy_inplace(self, selector: Union[npt.NDArray[bool], npt.NDArray[np.integer], slice]) -> None:
        """
        inplace row-selector (mask, fancy-index, or slice) applied
        """
        for v in self._items.values():
            # first item in tuple is the array
            arr = v[0]
            # preserve name when copying inplace
            name = arr.get_name()
            arr = arr[selector]
            arr.set_name(name)
            v[0] = arr

    # -----------------------------------------
    def copy(self, cols=None, deep=False):
        """
        Returns a shallow copy of the item container.
        cols list can be provided for specific selection.
        """

        newcontainer = ItemContainer()
        if cols is None:
            newcontainer._items = self._items.copy()
            for k, v in newcontainer._items.items():
                newcontainer._items[k] = v.copy()
        else:
            for k in cols:
                newcontainer._items[k] = self._items[k].copy()
        return newcontainer

    # -----------------------------------------
    def copy_apply(self, func, *args, cols=None):
        """
        Returns a copy of the itemcontainer, applying a function to items before swapping them out in the new ItemContainer object.
        Used in Dataset row masking.
        """

        newcontainer = ItemContainer()

        if cols is None:
            for k, v in self._items.items():
                # tuple copy
                v = v.copy()
                newcontainer._items[k] = v
                v[0] = func(v[0], *args)
        else:
            for k in cols:
                # tuple copy
                v = self._items[k].copy()
                newcontainer._items[k] = v
                v[0] = func(v[0], *args)

        return newcontainer

    # -----------------------------------------
    def apply(self, func, *args, cols=None):
        """
        Performs a possibly inplace operation on items in the itemcontainer
        """
        if cols is None:
            for v in self._items.values():
                func(v[0], *args)
        else:
            for k in cols:
                v = self._items[k]
                func(v[0], *args)

    # -----------------------------------------
    def __eq__(self, other):
        if isinstance(other, ItemContainer):
            return self._items == other._items
        return self._items == other

    def __ne__(self, other):
        if isinstance(other, ItemContainer):
            return self._items != other._items
        return self._items != other

    # def __del__(self):
    #    self._items.clear()                # eliminate cyclical references

    # -----------------------------------------
    def items_as_dict(self):
        """
        Return dictionary of items without attributes.
        """
        return {k: v[0] for k, v in self._items.items()}

    # -----------------------------------------
    def items_tolist(self):
        return [v[0] for v in self._items.values()]

    # -----------------------------------------
    def item_delete(self, key):
        del self._items[key]

    # -------------------------------------------------------
    def item_get_dict(self):
        """
        return the underlying dict

        values are stored in the first tuple, attributes in the second tuple
        """
        return self._items

    # -------------------------------------------------------
    def iter_values(self):
        """
        This will yield the full values in _items dict (lists with item, attribute)
        """
        for v in self._items.values():
            yield v

    # -------------------------------------------------------
    def item_get_value(self, key):
        """
        return the value for the given key

        NOTE: a good spot to put a counter for debugging
        """
        return self._items[key][0]

    # -------------------------------------------------------
    def item_get_values(self, keylist):
        """
        return list of value for the given key
        used for fast dataset slicing/copy with column selection
        """
        return [self.item_get_value(i) for i in keylist]

    # -------------------------------------------------------
    def item_set_value(self, key, value, attr=None):
        # check if already exists...
        temp = [value, attr]
        v = self._items.setdefault(key, temp)
        if v is not temp:
            v[0] = value

    def item_set_value_internal(self, key, value):
        # no checks, go to dict
        self._items[key] = value

    # -------------------------------------------------------
    def item_get_attribute(self, key, attrib_name, default=None):
        """
        Retrieve the value of the attribute previously assigned with `item_set_attribute`.

        Parameters
        ----------
        key:
            name of the item
        attrib_name:
            name of the attribute
        """
        item = self._items.get(key, None)
        if item is None:
            return None
        attrib = item[1]
        if attrib is None:
            return None
        return getattr(attrib, attrib_name, default)

    # -------------------------------------------------------
    def _set_attribute(self, item, name, value):
        attrib = item[1]
        if attrib is None:
            attrib = ItemAttribute()
        setattr(attrib, name, value)
        item[1] = attrib

    # -------------------------------------------------------
    def item_set_attribute(self, key, attrib_name, attrib_value):
        """
        Attach an attribute (name,value) pair to the item.

        Any valid dictionary name and any object can be assigned.

        Note: see `item_get_attribute` to retrieve.

        Parameters
        ----------
        key:
            name of the item
        attrib_name:
            name of the attribute
        attrib_value:
            value of the attribute
        """

        # check if already exists...
        if self.item_exists(key):
            self._set_attribute(self._items[key], attrib_name, attrib_value)
        else:
            raise KeyError(f"{key!r} does not already exist, thus cannot add attribute")

    # -------------------------------------------------------
    def item_get_len(self):
        return len(self._items)

    # -------------------------------------------------------
    def item_exists(self, item):
        return item in self._items

    # -------------------------------------------------------
    def get_dict_values(self):
        """
        Returns a tuple of items in the item dict. Each item is a list.
        """
        return tuple(self._items.values())

    # -------------------------------------------------------
    def item_replace_all(self, newdict, check_exists=True):
        """
        Replace the data for each item in the item dict. Original attributes
        will be retained.

        Parameters
        ----------
        newdict : dictionary of item names -> new item data (can also be a dataset)

        check_exists : if True, all newdict keys and old item keys will be compared to ensure a match
        """
        # for intenal routines, an existance check can often be skipped
        if check_exists:
            for k in newdict:
                if self.item_exists(k) is False:
                    raise ValueError(f"Item {k} not found in original item dictionary.")
            for k in self._items:
                if k not in newdict:
                    raise ValueError(f"Item {k} in original item dictionary not found in new items.")

        # replace the data, keep any attributes if set
        for k, v in newdict.items():
            self._items[k][0] = v

    # -------------------------------------------------------
    def item_rename(self, old, new):
        """
        Rename a single column.

        :param old: Current column name.
        :param new: New column name.
        :return: value portion of item that was renamed
        """
        if old == new:
            return None
        if old not in self._items:
            raise ValueError(
                f"Invalid column to rename: {old!r} cannot rename column that does not exit in itemcontainer."
            )
        if new in self._items:
            raise ValueError(f"Invalid column name: {new!r}; already exists in itemcontainer, cannot rename to it.")

        newdict = self._items.copy()

        return_val = None

        self._items.clear()
        for k, v in newdict.items():
            if k == old:
                k = new

                # return the value portion
                return_val = v[0]

            self._items[k] = v

        return return_val

    # -------------------------------------------------------
    def _get_move_cols(self, cols):
        """
        Possibly convert list/array/dictionary/string/index of items to move
        for item_move_to_front(), item_move_to_back()
        """
        if isinstance(cols, (str, bytes)):
            cols = [cols]
        elif isinstance(cols, (int, np.integer)):
            try:
                cols = [list(self._items.keys())[cols]]
            except:
                raise ValueError(f"Items could not be indexed by {cols}")

        if not isinstance(cols, (np.ndarray, list, tuple, dict)):
            raise TypeError(
                f"Item(s) to move must be list, tuple, ndarray, dictionary (keys), single unicode or byte string, or single index. Got {type(cols)}"
            )
        else:
            if len(cols) > len(self._items):
                raise ValueError(
                    f"Found {len(cols)} items to move to front - more than {len(self._items)} in container."
                )

        return cols

    # -------------------------------------------------------
    def item_move_to_front(self, cols):
        """
        Move single column or group of columns to front of list for iteration/indexing/display.
        Values of columns will remain unchanged.

        :param cols: list of column names to move.
        :return: None
        """
        cols = self._get_move_cols(cols)

        new_all_items = {}
        for cn in cols:
            if cn in self._items:
                new_all_items[cn] = self._items[cn]
            else:
                warnings.warn(f"Column {cn} not found. Could not move to front.")
        for cn in self._items:
            if cn not in new_all_items:
                new_all_items[cn] = self[cn]
        self._items = new_all_items

    # -------------------------------------------------------
    def item_move_to_back(self, cols):
        """
        Move single column or group of columns to front of list for iteration/indexing/display.
        Values of columns will remain unchanged.

        :param cols: list of column names to move.
        :return: None
        """
        cols = self._get_move_cols(cols)

        all_names = self._items
        for cn in cols:
            if cn in all_names:
                all_names[cn] = all_names.pop(cn)
            else:
                warnings.warn(f"Column {cn} not found. Could not move to back.")

    # -------------------------------------------------------
    def item_add_prefix(self, prefix):
        """
        inplace operation.
        adds prefix in front of existing item name

        faster than calling rename
        """
        newdict = self._items.copy()
        self._items.clear()
        for k, v in newdict.items():
            self._items[f"{prefix}{k}"] = v

    # -------------------------------------------------------
    def item_add_suffix(self, suffix):
        """
        inplace operation.
        adds suffix in front of existing item name

        faster than calling rename
        """
        newdict = self._items.copy()
        self._items.clear()
        for k, v in newdict.items():
            self._items[f"{k}{suffix}"] = v

    # --------------------------------------------------------
    def item_str_match(self, expression, flags=0):
        """
        Create a boolean mask vector for items whose names match the regex.
        NB Uses re.match(), not re.search().

        :param expression: regular expression
        :param flags: regex flags (from re module).
        :return: list array of bools (len ncols) which is true for columns which match the regex.
        """
        match_fun = re.compile(expression, flags=flags).match
        return [bool(match_fun(x)) for x in self._items.keys()]

    def item_str_replace(self, old, new, maxr=-1):
        """
        :param old: string to look for within individual names of columns
        :param new: string to replace old string in column names

        If an item name contains the old string, the old string will be replaced with the new one.
        If replacing the string conflicts with an existing item name, an error will be raised.

        returns True if column names were replaced
        """
        new_names = []
        replace_count = 0
        for item in self._items:
            r = item.replace(old, new, maxr)
            if r != item:
                replace_count += 1
                # prevent name conflict from overwriting existing column
                if r in self._items:
                    raise ValueError(f"Item {r} already existed, cannot make replacement in item name {item}.")
            new_names.append(r)

        # only do this if necessary
        if replace_count != 0:
            newdict = self._items.copy()
            self._items.clear()
            for i, v in enumerate(newdict.values()):
                self._items[new_names[i]] = v
            return True

        return False

    # -------------------------------------------------------
    def footer_get_value(self, key):
        return self.item_get_attribute(key, ATTRIBUTE_FOOTER)

    def footer_set_value(self, key, value):
        self.item_set_attribute(key, ATTRIBUTE_FOOTER, value)

    # --LABEL (LEFT FOR DISPLAY)-----------------------------
    # -------------------------------------------------------
    def label_as_dict(self):
        return self._tagged_as_dict(ATTRIBUTE_LABEL)

    def label_get_names(self):
        return self._tagged_get_names(ATTRIBUTE_LABEL)

    def label_set_names(self, listnames):
        self._tagged_set_names(listnames, ATTRIBUTE_LABEL)

    def label_remove(self):
        self._tagged_remove(ATTRIBUTE_LABEL)

    # --RIGHT FOR DISPLAY------------------------------------
    # -------------------------------------------------------
    def summary_as_dict(self):
        return self._tagged_as_dict(ATTRIBUTE_SUMMARY)

    def summary_get_names(self):
        return self._tagged_get_names(ATTRIBUTE_SUMMARY)

    def summary_set_names(self, listnames):
        self._tagged_set_names(listnames, ATTRIBUTE_SUMMARY)

    def summary_remove(self):
        self._tagged_remove(ATTRIBUTE_SUMMARY)

    # --GENERAL ATTRIBUTE FUNCTIONS--------------------------
    # -------------------------------------------------------
    def _tagged_get_names(self, attrname):
        """
        Returns a list of item names tagged with attrname in order.
        """
        tagged_names = []
        max, tagged_dict = self._tagged_get_dict_max(attrname)
        if max >= 0:
            for i in range(max + 1):
                if i in tagged_dict:
                    tagged_names.append(tagged_dict[i])
        return tagged_names

    def _tagged_set_names(self, listnames, attrname):
        """
        Removes existing items tagged with attrname.
        If items in listnames exist, they will be tagged with attrname.
        """
        if not isinstance(listnames, list):
            listnames = [listnames]
        self._tagged_remove(attrname)
        for i, tagged in enumerate(listnames):
            if self.item_exists(tagged):
                self._set_attribute(self._items[tagged], attrname, i)

    def _tagged_remove(self, attrname):
        """
        Removes existing items tagged with attrname.
        """
        for v in self._items.values():
            # get the attribute tuple
            attr = v[1]
            if attr is not None and hasattr(attr, attrname):
                delattr(attr, attrname)

    def _tagged_as_dict(self, attrname):
        """
        Returns dictionary of columns tagged with attrname.
        """
        return_dict = {}
        max, tagged_dict = self._tagged_get_dict_max(attrname)
        if max >= 0:
            for i in range(max + 1):
                if i in tagged_dict:
                    name = tagged_dict[i]
                    return_dict[name] = self.item_get_value(name)
        if len(return_dict) > 0:
            return return_dict
        return None

    def _tagged_get_dict_max(self, attrname):
        """
        Returns unordered dictionary of columns tagged with attrname, max value for order.
        """
        tagged_dict = {}
        max = -1
        for k, v in self._items.items():
            # get the attribute tuple
            attr = v[1]
            if attr is not None and hasattr(attr, attrname):
                val = getattr(attr, attrname)
                if val > max:
                    max = val
                tagged_dict[val] = k
        return max, tagged_dict



from typing import Optional, Union

__all__ = [
    "Multiset",
]

import warnings

import numpy as np

from .rt_dataset import Dataset
from .rt_display import DisplayDetect, DisplayString, DisplayTable
from .rt_enum import DS_DISPLAY_TYPES, ColHeader, DisplayDetectModes, TypeRegister
from .rt_misc import build_header_tuples
from .rt_sort_cache import SortCache
from .rt_struct import Struct


# =====================================================================================
class Multiset(Struct):
    """
    Multisets contain datasets and/or multisets where all contained dataset have the
    same number of rows. Multisets can provide a convenient namespace for closely
    related datasets, such as those loaded from a single HDF5 file or generated by
    an aggregation applied to a GroupBy object.

    The columns within contained datasets may be displayed in an interleaved way.
    Example: Assume Jan and Feb are two datasets with 3 columns each:

    * Jan: Run1, Run2, Run3
    * Feb: Run1, Run2, Run3

    A Multiset containing these datasets would display with a multi-line header:

    Run1        Run2        Run3

    Jan   Feb    Jan  Feb    Jan  Feb

    One can access the Run1 column in the Jan dataset with the syntax:
    ms.Jan.Run1

    Examples
    --------
    >>> ds=rt.Dataset({'somenans': [0., 1., 2., nan, 4., 5.], 'morestuff': ['A','B','C','D','E','F']})
    >>> ds2=rt.Dataset({'somenans': [0., 1., nan, 3., 4., 5.], 'morestuff':['H','I','J','K','L','M']})
    >>> ms=rt.Multiset({'test':ds, 'test2':ds2})
    >>> ms
          somenans      morestuff
    #   test   test2   test   test2
    -   ----   -----   ----   -----
    0   0.00    0.00   A      H
    1   1.00    1.00   B      I
    2   2.00     nan   C      J
    3    nan    3.00   D      K
    4   4.00    4.00   E      L
    5   5.00    5.00   F      M

    >>> ms['morestuff']
         morestuff
    #   test   test2
    -   ----   -----
    0   A      H
    1   B      I
    2   C      J
    3   D      K
    4   E      L
    5   F      M

    >>> ms['test']
    #   somenans   morestuff
    -   --------   ---------
    0       0.00   A
    1       1.00   B
    2       2.00   C
    3        nan   D
    4       4.00   E
    5       5.00   F

    >>> ms[[2,3],'somenans']
          somenans
    #   test   test2
    -   ----   -----
    0   2.00     nan
    1    nan    3.00

    >>> ms[[2,3],'morestuff']
         morestuff
    #   test   test2
    -   ----   -----
    0   C      J
    1   D      K

    >>> ms[[2,3],['morestuff','somenans']]
         morestuff       somenans
    #   test   test2   test   test2
    -   ----   -----   ----   -----
    0   C      J       2.00     nan
    1   D      K        nan    3.00
    """

    # ------------------------------------------------------------
    def __init__(self, input_value: Optional[Union[dict, "Multiset"]] = None):
        if input_value is None:
            input_value = dict()
        self._pre_init()
        if isinstance(input_value, dict):
            self._init_from_dict(input_value)
        elif isinstance(input_value, Multiset):
            raise TypeError(f"Use {self.__class__.__name__}.copy(deep=True/False) as a copy constructor.")
        else:
            raise TypeError(f"{self.__class__.__name__} must be initialized with a dictionary.")
        self._gbkeys = None
        self._post_init()
        self._sorted_row_idx = None

    # ------------------------------------------------------------
    def _init_from_dict(self, dictionary):
        # all __init__ paths funnel into this
        req_nrows = 0  # number of required rows for each contained dataset and multiset
        for k, v in dictionary.items():
            if k[0] != "_":
                # only allow Datasets and Multisets, if their number of rows are equal
                if isinstance(v, (TypeRegister.Dataset, TypeRegister.Multiset)):
                    if not req_nrows:
                        req_nrows = v._nrows
                    elif req_nrows != v._nrows:
                        raise ValueError(f"Number of rows in datasets do not match")
                    # add the dataset to this class
                    self.col_set_value(k, v)
                else:
                    raise TypeError(f"Cannot add {type(v).__name__} to a multiset")
        # fixup _nrows
        self._nrows = None if self.col_get_len() == 0 else req_nrows

    # ------------------------------------------------------------
    def _check_addtype(self, name, value):
        """called from subclassed Struct when a new item is added"""
        # currently only allow Datasets and Multisets
        if not isinstance(value, (TypeRegister.Dataset, TypeRegister.Multiset)):
            raise TypeError(f"Multiset only holds Datasets and Multisets not {type(value).__name__}.")
        # number of rows must match
        if self._nrows and self._nrows != value._nrows:
            raise ValueError(f"Number of rows do not match")
        # passed all the checks
        return value

    # ------------------------------------------------------------
    def multiget(self, index, deep=False):
        """
        Returns a new Multiset representing a one-level sub-sampling of the original.

        Parameters
        ----------
        index: An index specification.
        deep: bool, False
           If set to True will make deep copies

        Returns
        -------
        A new Multiset.
        """
        dd = {}
        if isinstance(index, slice):
            for _k, _v in self.items():
                temp = _v[index]
                if temp.get_ncols() > 0:
                    dd[_k] = temp.copy(deep=True) if deep else temp
        else:
            for _k, _v in self.items():
                cols = _v._mask_get_item(index, by_col_arg=True)
                lcols = [_c for _c in cols if _c in _v]
                if lcols:
                    dd[_k] = _v[lcols].copy(deep=True) if deep else _v[lcols]
        ms = Multiset(dd)
        return ms

    # ------------------------------------------------------------
    def __len__(self):
        # For Multiset we will return the number of rows for length
        return self._nrows

    def __setitem__(self, index, value):
        """
        Parameters
        ----------
        index: colspec
        value: A Dataset or Multiset

        Returns
        -------
        None

        Raises
        ------
        IndexError, TypeError, ValueError
        """
        if isinstance(index, (tuple, type(None))):
            raise IndexError(
                f"Can only index {self.__class__.__name__} as st[c], where c is colname, list of colnames or boolean mask."
            )

        # return: col_idx, row_idx, ncols, nrows, row_arg
        col_idx, _, ncols, _, _ = self._extract_indexing(index)
        if col_idx is None:
            col_idx = list(self.keys())
        if ncols == 0:
            return
        if ncols == 1:
            if not isinstance(value, (TypeRegister.Dataset, Multiset)):
                raise TypeError(f"Multiset can only hold Datasets or Multisets not {type(value)!r}.")
            if self._nrows and self._nrows != value._nrows:
                raise ValueError(f"Number of rows do not match")
            if not isinstance(col_idx, str):
                col_idx = col_idx[0]
            if self.col_exists(col_idx):
                self.__setattr__(col_idx, value)
            elif self.is_valid_colname(col_idx):
                self.__setattr__(col_idx, value)
            else:
                raise IndexError(f"Invalid column name: {col_idx!r}")
            self._nrows = value._nrows
        else:
            if not all([self.col_exists(_k) for _k in col_idx]):
                raise IndexError("If creating a new column can only do one at a time.")
            for _k1, _v2 in zip(col_idx, value):
                if not isinstance(_v2, (TypeRegister.Dataset, Multiset)):
                    raise TypeError(f"Multiset can only hold Datasets or Multisets.")
                if self._nrows and self._nrows != _v2._nrows:
                    raise ValueError(f"Number of rows do not match")
                setattr(self, _k1, _v2)
                self._nrows = _v2._nrows
        return

    # --------------------------------------------------------
    def _autocomplete(self) -> str:
        return f"Multiset{self.shape}"

    # --------------------------------------------------------
    def _copy(self, deep=False, rows=None, cols=None, base_index=0, cls=None):
        """
        Bracket indexing that returns a multiset will funnel into this routine.

        Parameters
        ----------
        deep : if True, perform a deep copy on column array
        rows : row mask
        cols : column mask
        base_index : used for head/tail slicing
        cls : class of return type, for subclass super() calls

        First argument must be deep.  Deep cannnot be set to None.  It must be True or False.
        """
        if cls is None:
            cls = type(self)

        ms = Multiset({})
        ms_locked = self.is_locked()
        if ms_locked:
            ms._unlock()

        if rows is None and cols is None:
            for dsetname, dset in self.items():
                if deep:
                    ms[dsetname] = dset.copy(deep=deep)
                else:
                    ms[dsetname] = dset
        else:
            for dsetname, dset in self.items():
                ms[dsetname] = dset._copy(deep=deep, rows=rows, cols=cols, base_index=base_index)
        if ms_locked:
            ms._lock()
        return ms

    def copy(self, deep=True):
        """
        Returns a shallow or deep copy of the multiset
        Defaults to a deepy copy.

        Parameters
        ----------
        deep: bool, default True
            Set to False for a shallow copy.
        """
        return self._copy(deep)

    @staticmethod
    def _depth_first(curobject, curdict, level, returnlist):
        """returns the max depth, list of dictionaries"""
        maxlevel = level

        if isinstance(curobject, TypeRegister.Multiset):
            # drill deeper
            for values in curdict.values():
                level = Multiset._depth_first(values, values.asdict(), level + 1, returnlist)
                if level > maxlevel:
                    maxlevel = level
        elif isinstance(curobject, TypeRegister.Dataset):
            # collect dictionary for dataset
            returnlist.append(curdict)
            # newlist=[(k,v) for k,v in curdict.items()]
            # print("appending", newlist)
            # returnlist.append(newlist)
        else:
            raise TypeError("_depth_first: error dont understand type", type)

        return maxlevel

    @staticmethod
    def _build_col_headers(rootobject, rootdict):
        """return a list of lists of ColHeaders"""
        tier1_list = list(rootdict.keys())
        numTier1 = len(tier1_list)
        if numTier1 == 0:
            print("Aborting - No datasets")
            return None
        if numTier1 <= 1:
            operation = tier1_list[0]
            ds = rootdict[operation]

            if rootobject._gbkeys is None:
                cols = list(ds.keys())
            else:
                # single column also needs to remove potential gbkeys from result
                cols = [k for k in ds.keys() if k not in rootobject._gbkeys]
            num_cols = len(cols)
            op_list = [operation] * num_cols

            # first line
            multiline_col = [build_header_tuples(cols, 1, 0), build_header_tuples(op_list, 1, 0)]
            # second line (same operation repeated)

            # allArrays = [ds.__getattribute__(c) for c in cols]
            allArrays = [ds.__getattr__(c) for c in cols]

            # return [build_header_tuples]
            return multiline_col, allArrays

        # depth first to count datasets and depth
        returnlist = []
        maxlevel = Multiset._depth_first(rootobject, rootdict, 0, returnlist)

        # count up total number of numpy arrays
        count = 0
        for i, dicts in enumerate(returnlist):
            count += len(dicts)
            # remove the groupby columns from display (will still be held in dataset)
            if rootobject._gbkeys is not None:
                for k in rootobject._gbkeys:
                    del returnlist[i][k]

        # print("level ", maxlevel, "count ", count)

        if count <= 0:
            print("Dataset count is zero or larger than level", count, maxlevel)
            return None

        # print("returndict", returnlist)

        # setup the list of ColHeaders, assume two lines
        multiline_col = [[], []]

        ##########################################
        #  Multiset/
        #       Jan
        #            Run1,Run2,Run3
        #       Feb
        #            Run1,Run2,Run3
        # ----------------------------------------------
        #
        #     Run1        Run2        Run3
        #  Jan   Feb    Jan  Feb    Jan  Feb
        # ----------------------------------------------
        #
        # ("Run1", 2, 0)             ("Run2", 2, 1)             ("Run3", 2,2)
        # ("Jan", 1, 0)("Feb",1,0) ("Jan", 1, 1)("Feb",1,1) ("Jan", 1, 2)("Feb",1,2)
        # numpy arrays: Jan.Run1/Feb.Run1/Jan.Run2/Feb.Run2/Jan.Run3/Feb.Run3
        #

        # allArrays will contain the numpy arrays to display
        allArrays = []

        colorgroup = 0
        # pass over all datasets until all arrays have been paired up
        for outerindex in range(numTier1):

            # search the next dataset for ANY array
            # then try to match the next datasets with the same key
            for k, v in returnlist[outerindex].items():

                # always add the first item
                allArrays.append(v)
                cell_span = 1
                newcol = ColHeader(tier1_list[outerindex], 1, colorgroup)
                multiline_col[1].append(newcol)

                # now scan other datasets for the same key -- pair up with sister items
                for innerindex in range(outerindex + 1, numTier1):
                    if k in returnlist[innerindex]:
                        # add this sister item
                        allArrays.append(returnlist[innerindex][k])
                        cell_span += 1
                        newcol = ColHeader(tier1_list[innerindex], 1, colorgroup)
                        multiline_col[1].append(newcol)

                        # remove this item so we do not include it in next outerindex pass
                        del returnlist[innerindex][k]

                # the first line is added after the second line is calculated
                newcol = ColHeader(k, cell_span, colorgroup)
                multiline_col[0].append(newcol)
                colorgroup += 1

        # for index in range(itemsPer):
        #    # loop over all tier1 objects
        #    tindex=0
        #    newcol = ColHeader(returnlist[tindex][index][0], len(tier1_list), index)
        #    multiline_col[0].append(newcol)
        #    #print("multiline: ", tindex, newcol)
        #    for name in tier1_list:
        #        #print("**", tindex, index, " --> ", tier1_list[tindex], returnlist[tindex][index])
        #        newcol=ColHeader(tier1_list[tindex], 1, index)
        #        multiline_col[1].append(newcol)
        #        #print("second multiline: ", index, newcol)
        #        allArrays.append(returnlist[tindex][index][1])
        #        tindex += 1

        #        #if (numDatasets ==2):
        #        ## figure out which columns are missing in each

        #        #missing1 = {key: None for key in ds_list[1]._all_items if key not in ds_list[0]._all_items}
        #        #missing2 = {key: None for key in ds_list[0]._all_items if key not in ds_list[1]._all_items}
        # print("First line", multiline_col[0])
        # print("Second line", multiline_col[1])
        # print("Arrays", allArrays)

        return multiline_col, allArrays

    # -------------------------------------------------------
    def _build_footers(self):
        """
        Still testing.
        TODO: speed up this python loop
        """
        footers = None
        num_ds = len(self)
        if num_ds > 0:
            try:
                first_footers = list(getattr(self[0], "_footers").values())[0]
                footers = []
                for f_idx, footer in enumerate(first_footers):
                    footers.append(footer)  # insert the first
                    for i in range(1, num_ds):  # insert each for the other datasets at the same index
                        current_footers = list(self[i]._footers.values())[0]
                        footers.append(current_footers[f_idx])
                footers = {"Total": TypeRegister.FastArray(footers)}

            except:
                pass
            # for i in num_ds
        return footers

    # -------------------------------------------------------
    def _last_row_stats(self):
        return f"[{self._nrows} rows x {self._ncols} columns]"

    def make_table(self, display_type):
        """
        Pretty-print code used by infrastructure.

        :param display_type: See rt.rt_enum.DS_DISPLAY_TYPES.
        :return: Display object or string.
        """
        ## BUGGY?
        multiline_col, allArrays = self._build_col_headers(self, self.asdict())
        footers = self._build_footers()

        table = DisplayTable()

        # _repr_html_
        if display_type == DS_DISPLAY_TYPES.HTML:
            DisplayDetect.Mode = DisplayDetectModes.HTML
            table._display_mode = DisplayDetectModes.HTML
            from_str = False

        # __repr__
        elif display_type == DS_DISPLAY_TYPES.REPR:
            # this repr will be called before _repr_html_ in jupyter
            if DisplayDetect.Mode == DisplayDetectModes.HTML:
                return ""
            from_str = False

        # __str__
        else:
            from_str = True

        if self._sort_display and self._sorted_row_idx is not None:
            sorted_row_idx = self._sorted_row_idx
        else:
            # remove all row sorts
            sorted_row_idx = None

        if self._gbkeys is None:
            keytups = [ColHeader("#", 1, 0)]
        else:
            keytups = [ColHeader(k, 1, 0) for k in self._gbkeys]
            # pad the other rows
            # TODO: take the multiline padding routine out of display to stop this from happening twice
            # padtup = [ColHeader('',len(keytups),0)]
            # for i, header_row in enumerate(multiline_col[:-1]):
            #    multiline_col[i] = padtup + header_row

            # add groupby columns to main data
            keycols = [*self._gbkeys.values()]
            allArrays = keycols + allArrays
        # fix bottom row here
        multiline_col[-1] = keytups + multiline_col[-1]

        result = table.build_result_table(
            multiline_col,
            allArrays,
            self._nrows,
            keys=self._gbkeys,
            sortkeys=self._col_sortlist,
            from_str=from_str,
            sorted_row_idx=sorted_row_idx,
            transpose_on=self._transpose_on,
        )

        return result + "\n\n" + self._last_row_stats()

    # -------------------------------------------------------
    def __str__(self):
        return self.make_table(DS_DISPLAY_TYPES.STR)

    # -------------------------------------------------------
    def __repr__(self):
        # this will be called before _repr_html_ in jupyter
        return self.make_table(DS_DISPLAY_TYPES.REPR)

    # -------------------------------------------------------
    def _repr_html_(self):
        return self.make_table(DS_DISPLAY_TYPES.HTML)

    # -------------------------------------------------------
    @property
    def dtypes(self):
        """
        Returns dictionary of dtype for each column.

        Returns
        -------
        dict
            Dictionary containing the dtype for each column in the Multiset.
        """
        return {colname: getattr(self, colname).dtype for colname in self.keys()}

    # -------------------------------------------------------
    def __getitem__(self, index):
        """
        Parameters
        ----------
        index : (rowspec, colspec) or colspec

        Returns
        -------
        the indexed row(s), cols(s), sub-dataset or single value

        Examples
        --------
        >>> ds=rt.Dataset({'somenans': [0., 1., 2., nan, 4., 5.]})
        >>> ds2=rt.Dataset({'somenans': [0., 1., nan, 3., 4., 5.]})
        >>> ms=rt.Multiset({'test':ds, 'test2':ds2})
        >>> ms[2,:]
              somenans
        #   test   test2
        -   ----   -----
        0   2.00     nan

        Raises
        ------
        IndexError
            When an invalid column name is supplied.
        TypeError
        """

        def single_array(col_idx, row_idx):
            # will either return dataset or return an error
            try:
                ds = self.col_get_value(col_idx)
            except Exception:
                # search and collect all sub columns
                ms = {}
                for dsetname, ds in self.items():
                    try:
                        # have the sub column collect the information
                        newds = ds[row_idx, col_idx] if row_idx is not None else ds[col_idx]
                        if isinstance(newds, np.ndarray):
                            # convert to a dataset
                            ms[dsetname] = Dataset({col_idx: newds})
                        else:
                            ms[dsetname] = newds
                    except Exception:
                        pass
                if len(ms) == 0:
                    raise IndexError(f"Could not find column or sub column named: {col_idx}")
                else:
                    return Multiset(ms)

            if row_idx is not None:
                # array indexing takes place early here
                return ds[row_idx, :]
            else:
                return ds

        if isinstance(index, (int, np.integer, slice)):
            # call into Struct to get col number
            return super().__getitem__(index)

        col_idx, row_idx, ncols, nrows, row_arg = self._extract_indexing(index)

        # check for a single string which selects a single column
        if isinstance(col_idx, str):
            return single_array(col_idx, row_idx)

        # if a single integer specified, make a list of one number for fancy column indexing
        if isinstance(row_arg, (int, np.integer)):
            row_idx = [row_arg]

        return self._copy(deep=False, rows=row_idx, cols=col_idx)

    # -------------------------------------------------------
    def flatten(self, horizontal=True, delimiter="_", dset_col_name="Column"):
        """
        Return a single dataset constructed by concatenating all of the datasets and flattened
        multisets contained within the multiset. Horizontal flattening will concatenate the datasets
        horizontally, prepending the dataset name to each dataset's column names.
        Vertical flattening requires the names and order of columns in each dataset to be
        identical, adding a single column to the returned dataset containing the name
        of the dataset from which each row derives.

        Parameters
        ----------
        horizontal: bool
            If True, concatenate the Datasets horizontally, otherwise vertically.
        delimiter: char
            The character used when joining dataset and column names to create unique names.
        dset_col_name: string
            For vertical flattening, the name for the column containing dataset names.

        Returns
        -------
        Dataset

        Raises
        ------
        ValueError
        """
        initial = True
        for name in self:
            elem = self[name]
            from_multiset = False
            # Recursively flatten any contained Multisets
            if isinstance(elem, Multiset):
                elem = elem.flatten(horizontal, delimiter, dset_col_name)
                from_multiset = True
            if initial:
                # Requirements are based on the first dataset encountered
                label_col_names = elem.label_get_names()  # Label column names, if any
                col_names = [n for n in elem.keys() if n not in ([dset_col_name] + label_col_names)]
            else:
                if label_col_names != elem.label_get_names():
                    raise ValueError(f"Label columns in {name!r} do not match")
            val_col_names = [n for n in elem.keys() if n not in label_col_names]
            if horizontal:
                if initial:
                    dsr = elem[label_col_names] if label_col_names else TypeRegister.Dataset()
                dsr = TypeRegister.Dataset.concat_columns([dsr, elem[val_col_names]], do_copy=False)
                for col_name in val_col_names:
                    dsr.col_rename(col_name, delimiter.join([name, col_name]))
            else:
                in_col_names = [n for n in elem.keys() if n not in ([dset_col_name] + label_col_names)]
                if col_names != in_col_names:
                    raise ValueError(f"Columns in {name!r} do not match")
                inds = elem[label_col_names] if label_col_names else TypeRegister.Dataset()
                if not from_multiset:
                    if dset_col_name in elem.keys():
                        raise ValueError(f"Column name for datasets {dset_col_name!r} already exists in {name!r}")
                    inds[dset_col_name] = [name] * elem.shape[0]
                else:
                    # Append multiset name to dataset name column
                    elem[dset_col_name] = [delimiter.join([name, cn.decode()]) for cn in elem[dset_col_name]]
                inds = TypeRegister.Dataset.concat_columns([inds, elem[val_col_names]], do_copy=False)
                dsr = inds if initial else TypeRegister.Dataset.concat_rows([dsr, inds])
                # Move the label column(s) to front, so they are there even with label removed
                dsr.col_move_to_front(label_col_names)
            initial = False
        dsr.label_set_names(label_col_names)
        return dsr

    ## -------------------------------------------------------
    # def sort_view(self, refdataset, by, ascending=True, kind='mergesort', na_position='last'):
    #    '''
    #    Sorts all columns in all datasets only when displayed.
    #    This routine is fast and does not change data underneath.

    #    See also: sort_inplace and sort_copy
    #    '''
    #    refds = self[refdataset]
    #    self._sorted_row_idx = refds._sort_lexsort(by, ascending)
    #    self._sort_display = True
    #    return self

    # -------------------------------------------------------
    def label_set_names(self, listnames):
        """Set which column names can be used as labels in display"""
        gbkeys = {}
        for ds in self.values():
            for colname in listnames:
                gbkeys[colname] = ds[colname]
            # check if we found all the names
            if len(gbkeys) == len(listnames):
                break
        if len(gbkeys) == 0:
            raise ValueError(f"Could not find any columns named {listnames}")
        # tell display which columns to move to left
        self._gbkeys = gbkeys

    # -------------------------------------------------------
    def label_fixup(self):
        """Auto scan for which column names can be used as labels in display"""
        gbkeys = {}
        for ds in self.values():
            listnames = ds.label_get_names()
            if len(listnames) > 0:
                for colname in listnames:
                    gbkeys[colname] = ds[colname]
                break
        if len(gbkeys) == 0:
            self._gbkeys = None
        else:
            self._gbkeys = gbkeys
        # allow chaining from this method
        return self

    # -------------------------------------------------------
    def cascade(self, funcname, *args, **kwargs):
        """
        Depth first calling of functions, often into a Dataset.
        For each Dataset in the Multiset, the function will be called with the `args` and `kwargs`.
        The return result is expected to be a Dataset which will then be added back into a new Multiset
        and returned to the called.

        Parameters
        ----------
        funcname: string or callable function

        Returns
        -------
        Multiset
        """
        if callable(funcname):
            func = funcname
        else:
            func = MultisetFunc.get(funcname, None)

        if func is not None:
            ms = Multiset({})
            for dsetname, dset in self.items():
                ms[dsetname] = func(dset, *args, **kwargs)
            return ms.label_fixup()
        raise ValueError(f"The function name passed {funcname} could not be found or is not callable")

    # -------------------------------------------------------
    # Copied over from Dataset and then cascaded down
    def apply(self, *args, **kwargs):
        return self.cascade("apply", *args, **kwargs)

    def apply_cols(self, *args, **kwargs):
        return self.cascade("apply_cols", *args, **kwargs)

    def apply_rows(self, *args, **kwargs):
        return self.cascade("apply_rows", *args, **kwargs)

    def sort_inplace(self, *args, **kwargs):
        return self.cascade("sort_inplace", *args, **kwargs)

    def sort_copy(self, *args, **kwargs):
        return self.cascade("sort_copy", *args, **kwargs)

    def fillna(self, *args, **kwargs):
        return self.cascade("fillna", *args, **kwargs)

    def all(self, *args, **kwargs):
        return self.cascade("all", *args, **kwargs)

    def any(self, *args, **kwargs):
        return self.cascade("any", *args, **kwargs)

    def keep(self, *args, **kwargs):
        return self.cascade("keep", *args, **kwargs)

    def trim(self, *args, **kwargs):
        return self.cascade("trim", *args, **kwargs)

    def pivot(self, *args, **kwargs):
        return self.cascade("pivot", *args, **kwargs)

    def quantile(self, *args, **kwargs):
        return self.cascade("quantile", *args, **kwargs)

    def describe(self, *args, **kwargs):
        return self.cascade("describe", *args, **kwargs)

    def sum(self, *args, **kwargs):
        return self.cascade("sum", *args, **kwargs)

    def mean(self, *args, **kwargs):
        return self.cascade("mean", *args, **kwargs)

    def std(self, *args, **kwargs):
        return self.cascade("std", *args, **kwargs)

    def var(self, *args, **kwargs):
        return self.cascade("var", *args, **kwargs)

    def min(self, *args, **kwargs):
        return self.cascade("min", *args, **kwargs)

    def max(self, *args, **kwargs):
        return self.cascade("max", *args, **kwargs)

    def nansum(self, *args, **kwargs):
        return self.cascade("nansum", *args, **kwargs)

    def nanmean(self, *args, **kwargs):
        return self.cascade("nanmean", *args, **kwargs)

    def nanstd(self, *args, **kwargs):
        return self.cascade("nanstd", *args, **kwargs)

    def nanvar(self, *args, **kwargs):
        return self.cascade("nanvar", *args, **kwargs)

    def nanmin(self, *args, **kwargs):
        return self.cascade("nanmin", *args, **kwargs)

    def nanmax(self, *args, **kwargs):
        return self.cascade("nanmax", *args, **kwargs)

    def abs(self, *args, **kwargs):
        return self.cascade("abs", *args, **kwargs)

    def astype(self, *args, **kwargs):
        return self.cascade("astype", *args, **kwargs)


MultisetFunc = {}
# -----------------------------------------------------------
def _FixupDocStrings():
    """
    Load all the member function of Dataset module
    Look up functions we want to copy, copy over docstring from Dataset to Multiset
    """
    import inspect
    import sys

    mymodule = sys.modules[__name__]
    functions_to_copy = [
        "apply",
        "apply_cols",
        "apply_rows",
        "sort_inplace",
        "sort_copy",
        "fillna",
        "all",
        "any",
        "keep",
        "trim",
        "pivot",
        "quantile",
        "describe",
        "sum",
        "mean",
        "std",
        "var",
        "min",
        "max",
        "nansum",
        "nanmean",
        "nanstd",
        "nanvar",
        "nanmin",
        "nanmax",
        "abs",
        "astype",
    ]

    all_dataset_functions = inspect.getmembers(Dataset, inspect.isfunction)
    # print("**all", all_dataset_functions)
    # now for each function that has an np flavor, copy over the doc strings
    for funcs in all_dataset_functions:
        funcname = funcs[0]
        if funcname in functions_to_copy:
            dsfunc = getattr(Dataset, funcname)
            msfunc = getattr(Multiset, funcname)

            # copy doc string
            msfunc.__doc__ = dsfunc.__doc__

            # add to lookup table
            MultisetFunc[funcname] = dsfunc


# ----------------------------------------------------------
# this is called when the module is loaded
_FixupDocStrings()

# keep as last line
TypeRegister.Multiset = Multiset



import os
import pathlib
import shutil
import sys
import unittest
from typing import Optional

import pytest
from numpy.random import default_rng

import riptable as rt
from riptable import *
from riptable.rt_enum import CategoryMode, SDSFlag
from riptable.rt_sds import SDSMakeDirsOn
from riptable.testing.array_assert import assert_array_or_cat_equal
from riptable.testing.randgen import create_test_dataset
from riptable.tests.test_utils import get_all_categorical_data
from riptable.Utils.rt_metadata import MetaData
from riptable.Utils.rt_testdata import load_test_data
from riptable.Utils.rt_testing import (
    assert_array_equal_,
    assert_categorical_equal,
    name,
)

# change to true since we write into /tests directory
SDSMakeDirsOn()

arr_types = [np.int8, np.uint8, np.int16, np.uint16, np.int32, np.uint32, np.int64, np.uint64, np.float32, np.float64]


def arr_eq(a, b):
    return bool(np.all(a == b))


def arr_all(a):
    return bool(np.all(a))


class SaveLoad_Test(unittest.TestCase):
    def setUp(self) -> None:
        # TODO enable SDS verbose logging for local SDS pytest runs if the verbose flag is passed to pytest
        # See the pytest doc for how to do this dynamically
        # https://docs.pytest.org/en/latest/example/simple.html#pass-different-values-to-a-test-function-depending-on-command-line-options
        # SDSVerboseOn()
        pass

    def test_save_arrays(self):
        pass

    def test_save_numeric(self):
        ds = Dataset({dt.__name__: arange(10, dtype=dt) for dt in arr_types})
        ds.save(r"riptable/tests/temp/tempsave")
        ds2 = Dataset.load(r"riptable/tests/temp/tempsave")

        # name, column order matches
        loadkeys = list(ds2.keys())
        for i, k in enumerate(ds.keys()):
            self.assertEqual(loadkeys[i], k)

        # dtype
        for i, c in enumerate(ds2.values()):
            self.assertEqual(arr_types[i], c.dtype)

        os.remove(r"riptable/tests/temp/tempsave.sds")

    def test_save_strings(self):
        uni = FastArray(["a", "b", "c", "d", "e"], unicode=True)
        b = FastArray(["a", "b", "c", "d", "e"])
        ds = Dataset({"unicode": uni, "bytes": b}, unicode=True)
        ds.save(r"riptable/tests/temp/tempsave")
        ds2 = Dataset.load(r"riptable/tests/temp/tempsave")

        self.assertTrue(ds2.unicode.dtype.char, "U")
        self.assertTrue(ds2.bytes.dtype.char, "S")

        os.remove(r"riptable/tests/temp/tempsave.sds")

    def test_save_categorical(self):
        # stringlike
        cstr = Categorical(["a", "b", "c", "d", "e"])

        # numeric
        cnum = Categorical(FastArray([1, 2, 3, 4, 5]), FastArray([10, 20, 30, 40, 50]))

        # enum
        cmap = Categorical([10, 20, 30, 40, 50], {10: "a", 20: "b", 30: "c", 40: "d", 50: "e"})

        # multikey
        cmk = Categorical([FastArray(["a", "b", "c", "d", "e"]), arange(5)])

        ds = Dataset({"cstr": cstr, "cnum": cnum, "cmap": cmap, "cmk": cmk})
        ds.save(r"riptable/tests/temp/tempsave")
        ds2 = Dataset.load(r"riptable/tests/temp/tempsave")

        self.assertEqual(ds2.cstr.category_mode, CategoryMode.StringArray)
        self.assertEqual(ds2.cnum.category_mode, CategoryMode.NumericArray)
        self.assertEqual(ds2.cmap.category_mode, CategoryMode.Dictionary)
        self.assertEqual(ds2.cmk.category_mode, CategoryMode.MultiKey)

        os.remove(r"riptable/tests/temp/tempsave.sds")

    def test_load_single_item(self):
        st = Struct(
            {"ds": Dataset({dt.__name__: arange(10, dtype=dt) for dt in arr_types}), "col1": arange(5), "num": 13}
        )

        st.save(r"riptable/tests/temp/tempsavestruct")
        ds = Struct.load(r"riptable/tests/temp/tempsavestruct", name="ds")

        for k, model_col in st.ds.items():
            self.assertTrue(bool(np.all(model_col == ds[k])))

        shutil.rmtree(r"riptable/tests/temp/tempsavestruct")

    def test_sharedmem_save(self):
        # need to write different routine to remove temporary save from linux shared memory
        if False:
            if sys.platform == "windows":
                ds = Dataset({"col_" + str(i): np.random.rand(5) for i in range(5)})
                ds.save(r"riptable/tests/temp/temp_save_shared", share="test_save_shared")

                # make sure the dataset is only shared in shared memory
                self.assertFalse(os.path.exists(r"riptable/tests/temp/temp_save_shared.sds"))

                # after loading, compare dataset to original
                ds2 = Dataset.load(r"riptable/tests/temp/temp_save_shared", share="test_save_shared")
                for k, v in ds2.items():
                    self.assertTrue(bool(np.all(v == ds[k])))

                # test with a garbage filepath
                ds.save(r"Z:::/riptable/tests/temp/temp_save_shared", share="test_save_shared")
                ds2 = Dataset.load(r"Y::::::/differentfilepath/samename/temp_save_shared", share="test_save_shared")
                for k, v in ds2.items():
                    self.assertTrue(bool(np.all(v == ds[k])))

    def test_sharedmem_load(self):
        ds = Dataset({"col_" + str(i): np.random.rand(5) for i in range(5)})
        ds.save(r"riptable/tests/temp/ds_to_file")

        # NOTE: shared memory on Windows requires admin privs and currently disabled
        if False:
            ds2 = load_sds(r"riptable/tests/temp/ds_to_file", share="test_load_shared")
            # remove file on disk
            os.remove(r"riptable/tests/temp/ds_to_file.sds")
            # load only from share
            ds3 = load_sds(r"riptable/tests/temp/ds_to_file", share="test_load_shared")

            for d in [ds, ds3]:
                for k, v in d.items():
                    self.assertTrue(bool(np.all(v == ds[k])))

    def test_uncompressed_save(self):
        ds = Dataset({"col_" + str(i): zeros(100_000, dtype=np.int32) for i in range(50)})
        ds.save(r"riptable/tests/temp/tempsave_compressed.sds")
        ds.save(r"riptable/tests/temp/tempsave_uncompressed.sds", compress=False)

        compsize = os.stat(r"riptable/tests/temp/tempsave_compressed.sds").st_size
        uncompsize = os.stat(r"riptable/tests/temp/tempsave_uncompressed.sds").st_size

        # make sure the file was smaller (this is full of zeros, so compression should be extreme)
        self.assertTrue(compsize < uncompsize)

        compds = Dataset.load(r"riptable/tests/temp/tempsave_compressed.sds")
        uncompds = Dataset.load(r"riptable/tests/temp/tempsave_uncompressed.sds")

        for ds2 in [compds, uncompds]:
            for k, v in ds2.items():
                self.assertTrue(bool(np.all(v == ds[k])))

        os.remove(r"riptable/tests/temp/tempsave_compressed.sds")
        os.remove(r"riptable/tests/temp/tempsave_uncompressed.sds")

    def test_meta_tuples(self):
        # test meta tuples for all fastarray / subclasses
        dtn = DateTimeNano(["2018-01-09", "2000-02-29", "2000-03-01", "2019-12-31"], from_tz="NYC")
        span = dtn.hour_span
        c = Categorical(["b", "a", "c", "a"])
        norm = arange(4)

        ds = Dataset({"norm": arange(4), "dtn": dtn, "Cat": Categorical(["b", "a", "c", "a"]), "span": dtn.hour_span})
        ds.save(r"riptable/tests/temp/tempsave_ds.sds")
        # only load the meta tuples
        correct_ds = [(b"norm", 3), (b"dtn", 3), (b"Cat", 1), (b"span", 3), (b"Cat!col_0", 2)]
        ds_tups = decompress_dataset_internal(r"riptable/tests/temp/tempsave_ds.sds")[0][2]
        for idx, correct in enumerate(correct_ds):
            result_tup = ds_tups[idx]
            self.assertEqual(correct[0], result_tup[0])
            self.assertEqual(correct[1], result_tup[1])

        os.remove(r"riptable/tests/temp/tempsave_ds.sds")

        st = Struct(ds.asdict())
        st.save(r"riptable/tests/temp/tempsave_st")
        correct_st = [(b"norm", 1), (b"dtn", 1), (b"Cat", 1), (b"span", 1), (b"Cat!col_0", 0)]
        st_tups = decompress_dataset_internal(r"riptable/tests/temp/tempsave_st.sds")[0][2]
        for idx, correct in enumerate(correct_st):
            result_tup = st_tups[idx]
            self.assertEqual(correct[0], result_tup[0])
            self.assertEqual(correct[1], result_tup[1])

        os.remove(r"riptable/tests/temp/tempsave_st.sds")

    def test_struct_no_arrays(self):
        ds1 = Dataset({"col_" + str(i): arange(5) for i in range(3)})
        ds2 = Dataset({"col_" + str(i): arange(5) for i in range(3)})
        ds3 = Dataset({"col_" + str(i): arange(5) for i in range(3)})

        st = Struct({"ds1": ds1, "ds2": ds2, "ds3": ds3})
        st.save(r"riptable/tests/temp/tempsave_st")
        st2 = Struct.load(r"riptable/tests/temp/tempsave_st")

        shutil.rmtree(r"riptable/tests/temp/tempsave_st")

    def test_corrupt_sds(self):
        f = open(r"riptable/tests/temp/garbage.sds", "w")
        f.close()
        with self.assertRaises(ValueError):
            ds = Dataset.load(r"riptable/tests/temp/garbage.sds")
        os.remove(r"riptable/tests/temp/garbage.sds")

    def test_scalar_overflow(self):
        large_val = 0xFFFFFFFFFFFFFFFFF
        large_val_arr = np.asarray([large_val])
        # will make an object when put in array
        self.assertEqual(large_val_arr.dtype.char, "O")

        st = Struct({"val": large_val})
        with self.assertWarns(UserWarning):
            st.save(r"riptable/tests/temp/tempsave_st")

        st2 = Struct.load(r"riptable/tests/temp/tempsave_st")
        val = st2.val
        self.assertTrue(isinstance(val, bytes))

        os.remove(r"riptable/tests/temp/tempsave_st.sds")
        # shutil.rmtree(r'riptable/tests/temp/tempsave_st')

    def test_meta_tuples_new(self):
        st = Struct(
            {
                "sc1": 1,
                "arr1": arange(5),
                "ds1": Dataset({"col1": arange(5)}),
                "cat1": Categorical(["a", "b", "c"]),
                "arr2": arange(5),
                "sc2": 2,
                "ds2": Struct({"test": 1}),
            }
        )
        correct_order = ["sc1", "arr1", "ds1", "cat1", "arr2", "sc2", "ds2"]
        correct_tuples = [
            (b"sc1", (SDSFlag.OriginalContainer | SDSFlag.Scalar)),
            (b"arr1", (SDSFlag.OriginalContainer)),
            (b"ds1", (SDSFlag.OriginalContainer | SDSFlag.Nested)),
            (b"cat1", (SDSFlag.OriginalContainer)),
            (b"arr2", (SDSFlag.OriginalContainer)),
            (b"sc2", (SDSFlag.OriginalContainer | SDSFlag.Scalar)),
            (b"ds2", (SDSFlag.OriginalContainer | SDSFlag.Nested)),
            (b"cat1!col_0", 0),
        ]

        st.save(r"riptable/tests/temp/tempsave_st")

        _, _, tups, _ = decompress_dataset_internal(r"riptable/tests/temp/tempsave_st/_root.sds")[0]
        for t, correct_t in zip(tups, correct_tuples):
            self.assertEqual(t[0], correct_t[0])
            self.assertEqual(t[1], correct_t[1])

        shutil.rmtree(r"riptable/tests/temp/tempsave_st")

    def test_load_extra_files(self):
        ds = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        st = Struct({"ds1": ds, "ds2": ds, "ds3": ds})

        st.save(r"riptable/tests/temp/tempsave_st")
        ds.save(r"riptable/tests/temp/tempsave_st/ds4")

        st2 = Struct.load(r"riptable/tests/temp/tempsave_st", include_all_sds=True)
        self.assertTrue("ds4" in st2)

        st2 = load_sds(r"riptable/tests/temp/tempsave_st", include_all_sds=True)
        self.assertTrue("ds4" in st2)

        shutil.rmtree(r"riptable/tests/temp/tempsave_st")

    def test_missing_item(self):
        ds = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        st = Struct({"ds1": ds, "ds2": ds, "ds3": ds})

        st.save(r"riptable/tests/temp/tempsave_st")
        os.remove(r"riptable/tests/temp/tempsave_st/ds2.sds")

        with self.assertWarns(UserWarning):
            st2 = load_sds(r"riptable/tests/temp/tempsave_st")
        self.assertTrue("ds2" not in st2)

        shutil.rmtree(r"riptable/tests/temp/tempsave_st")

    def test_single_items(self):
        arr = arange(5)
        save_sds(pathlib.PurePath(r"riptable/tests/temp/temparray"), arr)
        arr2 = load_sds(pathlib.PurePath(r"riptable/tests/temp/temparray"))
        self.assertTrue(bool(np.all(arr == arr2)))
        os.remove(pathlib.PurePath(r"riptable/tests/temp/temparray.sds"))

        cat = Categorical(["a", "a", "b", "c", "a"])
        save_sds(r"riptable/tests/temp/tempcat", cat)
        cat2 = load_sds(r"riptable/tests/temp/tempcat")
        self.assertTrue(isinstance(cat2, Categorical))
        self.assertTrue(bool(np.all(cat._fa == cat2._fa)))
        self.assertTrue(bool(np.all(cat.category_array == cat2.category_array)))
        os.remove(r"riptable/tests/temp/tempcat.sds")

        dtn = DateTimeNano(
            ["1992-02-01 12:34", "1995-05-12 12:34", "1956-02-07 12:34", "1959-12-30 12:34"], from_tz="NYC", to_tz="NYC"
        )
        save_sds(r"riptable/tests/temp/tempdtn", dtn)
        dtn2 = load_sds(r"riptable/tests/temp/tempdtn")
        self.assertTrue(isinstance(dtn2, DateTimeNano))
        self.assertTrue(bool(np.all(dtn._fa == dtn2._fa)))
        self.assertTrue(dtn._timezone._to_tz == dtn2._timezone._to_tz)
        os.remove(r"riptable/tests/temp/tempdtn.sds")

        arr = arange(24)
        arr = arr.reshape((2, 2, 2, 3))
        save_sds(r"riptable/tests/temp/temp4dims", arr)
        arr2 = load_sds(r"riptable/tests/temp/temp4dims")
        self.assertTrue(arr.dtype == arr2.dtype)
        self.assertTrue(arr.shape == arr2.shape)
        self.assertTrue(arr[1][1][1][1] == arr2[1][1][1][1])
        os.remove(r"riptable/tests/temp/temp4dims.sds")

    def test_shared_mem_errors(self):
        with self.assertRaises(ValueError):
            ds = load_sds(r"riptable/tests/temp/invalidfile.sds", share="invalidname")

        with self.assertRaises(ValueError):
            ds = load_sds_mem(r"riptable/tests/temp/invalidfile.sds", "invalidname")

    def test_include_dataset(self):
        ds = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        inc = ["col_1", "col_3"]
        ds.save(r"riptable/tests/temp/tempds")
        ds2 = load_sds(r"riptable/tests/temp/tempds", include=inc)
        self.assertEqual(ds2.shape[1], 2)
        self.assertTrue(list(ds2) == inc)
        self.assertTrue(isinstance(ds2, Dataset))
        os.remove(r"riptable/tests/temp/tempds.sds")

    def test_multistack_string_width(self):
        ds1 = Dataset({"strings": FastArray(["a", "b", "c", "d", "e"])})
        ds2 = Dataset({"strings": FastArray(["aa", "bb", "cc", "dd", "ee"])})
        files = [r"riptable/tests/temp/len1.sds", r"riptable/tests/temp/len2.sds"]
        ds1.save(files[0])
        ds2.save(files[1])

        arrays, _, _, _, _, _ = ds3 = rc.MultiStackFiles(files)
        stacked = arrays[0]
        self.assertTrue(isinstance(stacked, FastArray))
        self.assertEqual(stacked.itemsize, 2)
        self.assertEqual(stacked.dtype.name, "bytes16")

        for f in files:
            os.remove(f)

    def test_stack_missing_string(self):
        ds1 = Dataset({"col1": arange(5), "strings": FA(["a", "b", "c", "d", "e"], unicode=True)})
        ds2 = Dataset({"col1": arange(5)})
        files = [r"riptable/tests/temp/ds1.sds", r"riptable/tests/temp/ds2.sds"]
        ds1.save(files[0])
        ds2.save(files[1])

        ds3 = load_sds(files, stack=True)
        self.assertTrue(isinstance(ds3, Dataset))
        self.assertTrue(ds3._nrows, 10)
        self.assertTrue(ds3._ncols, 2)
        self.assertTrue(ds3.strings.dtype.char == "U")
        self.assertTrue(ds3.strings.itemsize == 4)

        for f in files:
            os.remove(f)

    # def test_stack_files(self):
    #    ds1 = Dataset({'col_'+str(i):np.random.rand(5) for i in range(5)})
    #    ds2 = Dataset({'col_'+str(i):np.random.rand(5) for i in range(5)})
    #    files = [ r'riptable/tests/temp/ds1.sds', r'riptable/tests/temp/ds2.sds' ]
    #    ds1.save(files[0])
    #    ds2.save(files[1])

    #    ds3 = load_sds(files, stack=True)
    #    self.assertTrue(isinstance(ds3, Dataset))
    #    self.assertEqual(ds3._nrows, 10)
    #    self.assertEqual(ds3._ncols, 5)

    #    top = ds3[:5,:]
    #    eq = top == ds1
    #    for col in eq.values():
    #        self.assertTrue(bool(np.all(col)))

    #    btm = ds3[5:,:]
    #    eq = btm == ds2
    #    for col in eq.values():
    #        self.assertTrue(bool(np.all(col)))

    #    for f in files:
    #        os.remove(f)

    def test_stack_files_include(self):
        ds1 = Dataset({"col_" + str(i): np.random.rand(5) for i in range(5)})
        ds2 = Dataset({"col_" + str(i): np.random.rand(5) for i in range(5)})
        files = [r"riptable/tests/temp/ds1.sds", r"riptable/tests/temp/ds2.sds"]
        ds1.save(files[0])
        ds2.save(files[1])

        inc = ["col_3", "col_4"]

        ds3 = load_sds(files, include=inc, stack=True)
        self.assertTrue(isinstance(ds3, Dataset))
        self.assertEqual(ds3._nrows, 10)
        self.assertEqual(ds3._ncols, 2)
        self.assertTrue(bool(np.all(inc == list(ds3))))

        for f in files:
            os.remove(f)

    # def test_stack_files_error(self):
    #    ds1 = Dataset({'nums':arange(5)})
    #    ds2 = Dataset({'nums':arange(5,dtype=np.int8)})
    #    files = [ r'riptable/tests/temp/ds1.sds', r'riptable/tests/temp/ds2.sds' ]
    #    ds1.save(files[0])
    #    ds2.save(files[1])

    #    with self.assertRaises(ValueError):
    #        ds3 = load_sds(files, stack=True)

    #    with self.assertRaises(TypeError):
    #        ds3 = load_sds( [ r'riptable/tests/temp/ds1.sds', r'riptable/tests/temp' ], stack=True )

    #    for f in files:
    #        os.remove(f)

    def test_stack_dir(self):
        ds1 = Dataset({"col_" + str(i): np.random.rand(5) for i in range(5)})
        ds2 = Dataset({"col_" + str(i): np.random.rand(5) for i in range(5)})
        files1 = [r"riptable/tests/temp/dir1/ds1.sds", r"riptable/tests/temp/dir2/ds1.sds"]
        files2 = [r"riptable/tests/temp/dir1/ds2.sds", r"riptable/tests/temp/dir2/ds2.sds"]
        for f in files1:
            ds1.save(f)
        for f in files2:
            ds2.save(f)

        dirs = [r"riptable/tests/temp/dir1", r"riptable/tests/temp/dir2"]
        inc = ["ds1", "ds2"]

        st = load_sds(dirs, include=inc, stack=True)

        self.assertTrue(isinstance(st, Struct))
        self.assertTrue(bool(np.all(inc == list(st))))
        for v in st.values():
            self.assertTrue(isinstance(v, Dataset))
            self.assertEqual(v._nrows, 10)
            self.assertEqual(v._ncols, 5)

        for d in dirs:
            shutil.rmtree(d)

    def test_stack_dir_error(self):
        ds1 = Dataset({"nums": arange(5)})
        ds2 = Dataset({"nums": arange(5, dtype=np.int8)})
        files = [r"riptable/tests/temp/ds1.sds", r"riptable/tests/temp/ds2.sds"]
        ds1.save(files[0])
        ds2.save(files[1])

        with self.assertRaises(ValueError):
            _ = load_sds([r"riptable/tests/temp"], stack=True)

        for f in files:
            os.remove(f)

    def test_stack_upcast(self):

        correct = tile(arange(5), 2)

        for dt in arr_types:
            ds = Dataset({"nums": arange(5, dtype=dt)})
            ds.save(r"riptable/tests/temp/upcast/" + np.dtype(dt).name)

        int_dt = [np.int8, np.int16, np.int32, np.int64]
        uint_dt = [np.uint8, np.uint16, np.uint32, np.uint64]
        flt_dt = [np.float32, np.float64]

        # float64
        flt64 = r"riptable/tests/temp/upcast/float64"
        for dt in int_dt + uint_dt + flt_dt:
            files = [flt64, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(
                bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: float64 and {np.dtype(dt).name}"
            )
            self.assertTrue(ds.nums.dtype == np.float64, msg=f"Failure in stacking: float64 and {np.dtype(dt).name}")

        # float32
        flt32 = r"riptable/tests/temp/upcast/float32"
        for dt in int_dt[:2] + uint_dt[:2]:
            files = [flt32, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(
                bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: float32 and {np.dtype(dt).name}"
            )
            self.assertTrue(ds.nums.dtype == np.float32, msg=f"Failure in stacking: float32 and {np.dtype(dt).name}")
        for dt in int_dt[2:] + uint_dt[2:]:
            files = [flt32, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(
                bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: float32 and {np.dtype(dt).name}"
            )
            self.assertTrue(ds.nums.dtype == np.float64, msg=f"Failure in stacking: float32 and {np.dtype(dt).name}")
        for dt in [int_dt[-1], uint_dt[-1], flt_dt[-1]]:
            files = [flt32, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(
                bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: float32 and {np.dtype(dt).name}"
            )
            self.assertTrue(ds.nums.dtype == np.float64, msg=f"Failure in stacking: float32 and {np.dtype(dt).name}")

        # int64
        i64 = r"riptable/tests/temp/upcast/int64"
        for dt in int_dt + uint_dt[:-1]:
            files = [i64, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: int64 and {np.dtype(dt).name}")
            self.assertTrue(ds.nums.dtype == np.int64, msg=f"Failure in stacking: int64 and {np.dtype(dt).name}")
        for dt in flt_dt:
            files = [i64, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: int64 and {np.dtype(dt).name}")
            self.assertTrue(ds.nums.dtype == np.float64, msg=f"Failure in stacking: int64 and {np.dtype(dt).name}")

        # int32
        i32 = r"riptable/tests/temp/upcast/int32"
        for dt in int_dt[:-1] + uint_dt[:2]:
            files = [i32, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: int32 and {np.dtype(dt).name}")
            self.assertTrue(ds.nums.dtype == np.int32, msg=f"Failure in stacking: int32 and {np.dtype(dt).name}")
        for dt in flt_dt:
            files = [i32, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: int32 and {np.dtype(dt).name}")
            self.assertTrue(ds.nums.dtype == np.float64, msg=f"Failure in stacking: int32 and {np.dtype(dt).name}")

        # int16
        i16 = r"riptable/tests/temp/upcast/int16"
        for dt in int_dt[:2] + uint_dt[:1]:
            files = [i16, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: int16 and {np.dtype(dt).name}")
            self.assertTrue(ds.nums.dtype == np.int16, msg=f"Failure in stacking: int16 and {np.dtype(dt).name}")
        for dt in flt_dt:
            files = [i16, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: int16 and {np.dtype(dt).name}")
            self.assertTrue(ds.nums.dtype == dt, msg=f"Failure in stacking: int16 and {np.dtype(dt).name}")

        # int8
        i8 = r"riptable/tests/temp/upcast/int8"
        for dt in int_dt + flt_dt:
            files = [i8, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: int8 and {np.dtype(dt).name}")
            self.assertTrue(ds.nums.dtype == dt, msg=f"Failure in stacking: int8 and {np.dtype(dt).name}")

        # uint64
        ui64 = r"riptable/tests/temp/upcast/uint64"
        for dt in uint_dt:
            files = [ui64, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(
                bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: uint64 and {np.dtype(dt).name}"
            )
            self.assertTrue(ds.nums.dtype == np.uint64, msg=f"Failure in stacking: uint64 and {np.dtype(dt).name}")
        for dt in int_dt[:-1]:
            files = [ui64, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(
                bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: uint64 and {np.dtype(dt).name}"
            )
            self.assertTrue(ds.nums.dtype == np.int64, msg=f"Failure in stacking: uint64 and {np.dtype(dt).name}")

        # uint32
        ui32 = r"riptable/tests/temp/upcast/uint32"
        for dt in int_dt[:-2]:
            files = [ui32, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(
                bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: uint32 and {np.dtype(dt).name}"
            )
            self.assertTrue(ds.nums.dtype == np.int64, msg=f"Failure in stacking: uint32 and {np.dtype(dt).name}")
        for dt in uint_dt[:-1]:
            files = [ui32, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(
                bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: uint32 and {np.dtype(dt).name}"
            )
            self.assertTrue(ds.nums.dtype == np.uint32, msg=f"Failure in stacking: uint32 and {np.dtype(dt).name}")

        # uint16
        ui16 = r"riptable/tests/temp/upcast/uint16"
        for dt in int_dt[:-3]:
            files = [ui16, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(
                bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: uint16 and {np.dtype(dt).name}"
            )
            self.assertTrue(ds.nums.dtype == np.int32, msg=f"Failure in stacking: uint16 and {np.dtype(dt).name}")
        for dt in uint_dt[:2]:
            files = [ui16, r"riptable/tests/temp/upcast/" + np.dtype(dt).name]
            ds = load_sds(files, stack=True)
            self.assertTrue(isinstance(ds, Dataset))
            self.assertTrue(
                bool(np.all(ds.nums == correct)), msg=f"Failure in stacking: uint16 and {np.dtype(dt).name}"
            )
            self.assertTrue(ds.nums.dtype == np.uint16, msg=f"Failure in stacking: uint16 and {np.dtype(dt).name}")

        # int + uint to larger itemsize
        # 32 -> 64
        files = [i32, ui32]
        ds = load_sds(files, stack=True)
        self.assertTrue(isinstance(ds, Dataset))
        self.assertTrue(bool(np.all(ds.nums == correct)), msg=f"Failed to upcast int32/uint32 -> int64")
        self.assertTrue(ds.nums.dtype == np.int64, msg="Failed to upcast int32/uint32 -> int64")

        # 16 -> 32
        files = [i16, ui16]
        ds = load_sds(files, stack=True)
        self.assertTrue(isinstance(ds, Dataset))
        self.assertTrue(bool(np.all(ds.nums == correct)), msg=f"Failed to upcast int16/uint16 -> int32")
        self.assertTrue(ds.nums.dtype == np.int32, msg="Failed to upcast int16/uint16 -> int32")

        ui8 = r"riptable/tests/temp/upcast/uint8"

        # 8 -> 16
        files = [i8, ui8]
        ds = load_sds(files, stack=True)
        self.assertTrue(isinstance(ds, Dataset))
        self.assertTrue(bool(np.all(ds.nums == correct)), msg=f"Failed to upcast int8/uint8 -> int16")
        self.assertTrue(ds.nums.dtype == np.int16, msg="Failed to upcast int8/uint8 -> int16")

        shutil.rmtree(r"riptable/tests/temp/upcast")

    def test_save_classname(self):
        files = [
            r"riptable/tests/temp/tempsave/cat",
            r"riptable/tests/temp/tempsave/dtn",
            r"riptable/tests/temp/tempsave/ts",
            r"riptable/tests/temp/tempsave/d",
            r"riptable/tests/temp/tempsave/dspan",
        ]
        c = Categorical(["a", "a", "b", "c", "a"])
        c.save(files[0])
        dtn = DateTimeNano.random(5)
        dtn.save(files[1])
        ts = TimeSpan(arange(100))
        ts.save(files[2])
        d = Date.range(20190202, 20191231)
        d.save(files[3])
        dspan = DateSpan(arange(100))
        dspan.save(files[4])

        types = [Categorical, DateTimeNano, TimeSpan, Date, DateSpan]
        typenames = [cls.__name__ for cls in types]

        # pull all the single items, compare typenames
        for i, t_name in enumerate(typenames):
            meta = MetaData(decompress_dataset_internal(files[i])[0][0])
            meta = MetaData(meta["item_meta"][0])
            self.assertEqual(meta["classname"], t_name)

        files = [
            r"riptable/tests/temp/tempsave/ds",
            r"riptable/tests/temp/tempsave/st",
            r"riptable/tests/temp/tempsave/pds",
        ]

        # test containers
        ds = Dataset({dt.__name__: arange(10, dtype=dt) for dt in arr_types})
        ds.save(files[0])
        st = Struct({"singlearr": arange(5)})
        st.save(files[1])
        pds = PDataset([ds, ds])
        pds.save(files[2])

        typenames = ["Dataset", "Struct", "PDataset"]
        for i, t_name in enumerate(typenames):
            meta = MetaData(decompress_dataset_internal(files[i])[0][0])
            self.assertEqual(typenames[i], meta["classname"])

        shutil.rmtree(r"riptable/tests/temp/tempsave")

    def test_load_list(self):
        files = [
            pathlib.PurePath(x)
            for x in [
                r"riptable/tests/temp/tempsave/ds",
                r"riptable/tests/temp/tempsave/st",
                r"riptable/tests/temp/tempsave/cat",
                r"riptable/tests/temp/tempsave/fa",
                r"riptable/tests/temp/tempsave/dtn",
                r"riptable/tests/temp/tempsave/ts",
                r"riptable/tests/temp/tempsave/d",
                r"riptable/tests/temp/tempsave/dspan",
            ]
        ]

        ds = Dataset({dt.__name__: arange(10, dtype=dt) for dt in arr_types})
        ds.save(files[0])
        st = Struct({"ds1": ds, "ds2": ds})
        st.save(files[1])
        c = Categorical(["a", "a", "b", "c", "a"])
        c.save(files[2])
        a = arange(1000)
        a.save(files[3])
        dtn = DateTimeNano.random(5)
        dtn.save(files[4])
        ts = TimeSpan(arange(100))
        ts.save(files[5])
        d = Date.range(20190202, 20191231)
        d.save(files[6])
        dspan = DateSpan(arange(100))
        dspan.save(files[7])

        reload = load_sds(files)
        types = [Dataset, Struct, Categorical, FastArray, DateTimeNano, TimeSpan, Date, DateSpan]

        for i, item in enumerate(reload):
            self.assertTrue(isinstance(item, types[i]), msg=f"expected {types[i]} got {type(item)}")

        shutil.rmtree(r"riptable/tests/temp/tempsave")

    def test_load_list_include(self):
        ds = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        ds.save(r"riptable/tests/temp/tempsave/ds1")
        ds.save(r"riptable/tests/temp/tempsave/ds2")
        files = [
            pathlib.PurePath(x) for x in [r"riptable/tests/temp/tempsave/ds1", r"riptable/tests/temp/tempsave/ds2"]
        ]
        inc = ["col_2", "col_3"]
        reload = load_sds(files, include=inc)
        for d in reload:
            self.assertTrue(isinstance(d, Dataset))
            self.assertTrue(bool(np.all(list(d) == inc)))

        shutil.rmtree(r"riptable/tests/temp/tempsave")

    def test_fix_strides(self):
        arr = arange(20, dtype=np.int32)
        arr = arr[::2]
        self.assertEqual(arr.strides[0], 8)
        arr.save(r"riptable/tests/temp/arr")

        arr2 = load_sds(r"riptable/tests/temp/arr")
        self.assertEqual(arr2.strides[0], 4, msg=f"{arr2}")

        self.assertTrue(bool(np.all(arr == arr2)))

        os.remove(r"riptable/tests/temp/arr.sds")

    def test_stack_fa_subclass(self):
        ds = Dataset(
            {
                "CAT": Categorical(np.random.choice(["a", "b", "c"], 1000)),
                "DTN": DateTimeNano.random(1000),
                "DATE": Date(np.random.randint(15000, 20000, 1000)),
                "TSPAN": TimeSpan(np.random.randint(0, 1_000_000_000 * 60 * 60 * 24, 1000, dtype=np.int64)),
                "DSPAN": DateSpan(np.random.randint(0, 365, 1000)),
            }
        )
        files = [r"riptable/tests/temp/ds" + str(i) + ".sds" for i in range(3)]
        for f in files:
            ds.save(f)

        pds = load_sds(files, stack=True)
        self.assertTrue(isinstance(pds, PDataset))
        for k, v in pds.items():
            self.assertEqual(type(v), type(ds[k]))

        for f in files:
            os.remove(f)

    def test_stack_multikey(self):
        paths = [r"riptable/tests/temp/ds1.sds", r"riptable/tests/temp/ds2.sds"]

        a_strings = FA(["b", "a", "c", "a"])
        a_nums = FA([1, 1, 3, 1])
        mkcat_a = Cat([a_strings, a_nums])
        single_a = Cat(FA(["b1", "a1", "c3", "a1"]))

        b_strings = FA(["a", "b", "a", "c"])
        b_nums = FA([1, 2, 1, 3])
        single_b = Cat(FA(["a1", "b2", "a1", "c3"]))
        mkcat_b = Cat([b_strings, b_nums])

        ds1 = Dataset({"mkcat": mkcat_a, "singlecat": single_a})
        ds1.save(paths[0])

        ds2 = Dataset({"mkcat": mkcat_b, "singlecat": single_b})
        ds2.save(paths[1])

        ds3 = load_sds(paths, stack=True)

        self.assertTrue(isinstance(ds3, PDataset))

        correct_single = FA(["b1", "a1", "c3", "a1", "a1", "b2", "a1", "c3"])
        self.assertTrue(arr_eq(correct_single, ds3.singlecat.expand_array))

        correct_mk_ikey = FastArray([1, 2, 3, 2, 2, 4, 2, 3])
        self.assertTrue(arr_eq(correct_mk_ikey, ds3.mkcat._fa))

        correct_mk_strings = FA(["b", "a", "c", "b"])
        correct_mk_nums = FA([1, 1, 3, 2])
        mk_cols = [correct_mk_strings, correct_mk_nums]
        result_cols = list(ds3.mkcat.category_dict.values())
        for i, col in enumerate(mk_cols):
            self.assertTrue(arr_eq(result_cols[i], col))

        for p in paths:
            os.remove(p)

    # def test_include_dataset_shared(self):
    #    ds = Dataset({'col_'+str(i):arange(5) for i in range(5)})
    #    inc = ['col_1', 'col_3']
    #    ds.save(r'riptable/tests/temp/tempds', share='testshare1')
    #    ds2 = load_sds(r'riptable/tests/temp/tempds', share='testshare1', include=inc)
    #    self.assertEqual(len(ds2), 2)
    #    self.assertTrue(list(ds2)==inc)
    #    self.assertTrue(isinstance(ds2, Dataset))
    #    with self.assertRaises(FileNotFoundError):
    #        os.remove(r'riptable/tests/temp/tempds.sds')

    # include keyword works differently with multi-file load
    # def test_include_struct(self):
    #    ds = Dataset({'col_'+str(i):arange(5) for i in range(5)})
    #    st = Struct({'ds1':ds, 'ds2':ds, 'ds3':ds, 'arr':arange(5)})
    #    inc = ['ds1','ds3','col_0','col_1','col_2','col_3','col_4']
    #    st.save(r'riptable/tests/temp/temp_st')
    #    st2 = load_sds(r'riptable/tests/temp/temp_st', include=inc)
    #    self.assertEqual(len(st2), 2)
    #    self.assertTrue(list(st2)==inc)
    #    self.assertTrue(isinstance(st2, Struct))
    #    self.assertFalse(isinstance(st2, Dataset))
    #    self.assertTrue(list(st.ds1) == list(st2.ds1))

    #    inc = ['ds2', 'arr','col_0','col_1','col_2','col_3','col_4']
    #    st2 = load_sds(r'riptable/tests/temp/temp_st', include=inc)
    #    self.assertEqual(len(st2), 2)
    #    self.assertTrue(list(st2)==inc)
    #    self.assertTrue(isinstance(st2, Struct))
    #    self.assertFalse(isinstance(st2, Dataset))
    #    self.assertTrue(list(st.ds2) == list(st2.ds2))

    #    shutil.rmtree(r'riptable/tests/temp/temp_st')

    # def test_include_struct_shared(self):
    #    ds = Dataset({'col_'+str(i):arange(5) for i in range(5)})
    #    st = Struct({'ds1':ds, 'ds2':ds, 'ds3':ds, 'arr':arange(5)})
    #    inc = ['ds1','ds3']
    #    st.save(r'riptable/tests/temp/temp_st', share='testshare2')
    #    st2 = load_sds(r'riptable/tests/temp/temp_st', share='testshare2', include=inc)
    #    self.assertEqual(len(st2), 2)
    #    self.assertTrue(list(st2)==inc)
    #    self.assertTrue(isinstance(st2, Struct))
    #    self.assertFalse(isinstance(st2, Dataset))
    #    self.assertTrue(list(st.ds1) == list(st2.ds1))

    #    inc = ['ds2', 'arr']
    #    st2 = load_sds(r'riptable/tests/temp/temp_st', share='testshare2', include=inc)
    #    self.assertEqual(len(st2), 2)
    #    self.assertTrue(list(st2)==inc)
    #    self.assertTrue(isinstance(st2, Struct))
    #    self.assertFalse(isinstance(st2, Dataset))
    #    self.assertTrue(list(st.ds2) == list(st2.ds2))
    #    with self.assertRaises(FileNotFoundError):
    #        shutil.rmtree(r'riptable/tests/temp/temp_st')

    def test_include_categorical(self):
        ds = Dataset({"arr": arange(5), "cat_col": Categorical(["a", "a", "b", "c", "a"])})
        ds.save(r"riptable/tests/temp/temp_cat")

        ds2 = load_sds(r"riptable/tests/temp/temp_cat", include=["cat_col"])
        self.assertTrue(list(ds2) == ["cat_col"])
        self.assertTrue(isinstance(ds2.cat_col, Categorical))
        self.assertTrue(bool(np.all(ds.cat_col == ds2.cat_col)))

        os.remove(r"riptable/tests/temp/temp_cat.sds")

    def test_include_categorical_multi(self):
        ds = Dataset({"arr": arange(5), "cat_col": Categorical([np.array(["a", "a", "b", "c", "a"]), arange(5)])})
        ds.save(r"riptable/tests/temp/temp_cat")

        ds2 = load_sds(r"riptable/tests/temp/temp_cat", include=["cat_col"])
        self.assertTrue(list(ds2) == ["cat_col"])
        self.assertTrue(isinstance(ds2.cat_col, Categorical))

        os.remove(r"riptable/tests/temp/temp_cat.sds")

    def test_sections(self):
        f = r"riptable/tests/temp/temp_cat.sds"
        ds = Dataset({"arr": arange(5), "cat_col": Categorical([np.array(["a", "a", "b", "c", "a"]), arange(5)])})
        for i in range(3):
            ds.save(f, append=str(i))
            ds.arr += 5
        ds2 = load_sds(f, stack=True)
        self.assertTrue(np.all(ds2.arr == arange(15)))

        # now add the section later (save first with no section name)
        ds = Dataset({"arr": arange(5), "cat_col": Categorical([np.array(["a", "a", "b", "c", "a"]), arange(5)])})
        ds.save(f)
        # remove catcol, change dataset length
        ds = Dataset({"arr": arange(15) + 5})
        ds.save(f, append="test")
        ds.arr += 15
        ds.save(f, append="test")

        # now read the data vertically using stack=True
        # can it auto gap fill a categorical?
        ds2 = load_sds(f, stack=True)
        self.assertTrue(np.all(ds2.arr == arange(35)))

        ds2 = load_sds(f, include=["cat_col"], stack=True)
        self.assertTrue(list(ds2) == ["cat_col"])
        self.assertTrue(isinstance(ds2.cat_col, Categorical))
        # make sure it dropped arr
        self.assertFalse(list(ds2) == ["arr"])

        # now do multiple files that have append section
        ds2 = load_sds([f, f], stack=True)
        # it should have stacked 6 sections total
        self.assertTrue(np.all(ds2.arr == hstack([np.arange(35), np.arange(35)])))

        # now read the data horizontal by removing stack=True
        ds2 = load_sds(f, include="arr", stack=False)
        # make sure there are 3 datasets
        self.assertTrue(len(ds2) == 3, msg=f"got {ds2}")

        # make sure the first dataset is the first dataset we saved
        self.assertTrue(np.all(ds2[0].arr == arange(5)))

        os.remove(f)

    def test_fix_path(self):
        ds = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        st = Struct({"ds1": ds, "ds2": ds, "ds3": ds, "arr": arange(5)})
        st.save(r"riptable/tests/temp/save_st_dir")
        # auto remove .sds extension
        st2 = Struct.load(r"riptable/tests/temp/save_st_dir.sds")
        self.assertTrue(isinstance(st2, Struct))
        st2 = load_sds(r"riptable/tests/temp/save_st_dir.sds")
        self.assertTrue(isinstance(st2, Struct))
        shutil.rmtree(r"riptable/tests/temp/save_st_dir")

        # auto add .sds extension
        ds.save(r"riptable/tests/temp/temp_save_ds")
        ds2 = Dataset.load(r"riptable/tests/temp/temp_save_ds")
        self.assertTrue(isinstance(ds2, Dataset))
        ds2 = load_sds(r"riptable/tests/temp/temp_save_ds")
        self.assertTrue(isinstance(ds2, Dataset))
        os.remove(r"riptable/tests/temp/temp_save_ds.sds")

        # not file or directory (.sds added)
        with self.assertRaises(ValueError):
            ds = Dataset.load(r"riptable/tests/temp/tempgarbagenamewithext.sds")
        with self.assertRaises(ValueError):
            st = Struct.load(r"riptable/tests/temp/tempgarbagenamewithext.sds")

        # not file or directory (without .sds)
        with self.assertRaises(ValueError):
            ds = Dataset.load(r"riptable/tests/temp/tempgarbagenamewihtoutext")
        with self.assertRaises(ValueError):
            ds = Struct.load(r"riptable/tests/temp/tempgarbagenamewihtoutext")

        # st.save(r'riptable/tests/temp/struct_with_ext.sds')
        # self.assertTrue(os.path.isdir(r'riptable/tests/temp/struct_with_ext'))
        # shutil.rmtree(r'riptable/tests/temp/struct_with_ext')

    def test_double_nested(self):
        st = Struct({"nested": Struct({"ds": Dataset({"arr": arange(5)})})})
        st.save(r"riptable/tests/temp/save_st")
        st2 = load_sds(r"riptable/tests/temp/save_st")
        self.assertTrue(isinstance(st2, Struct))
        # struct, not dictionary
        self.assertTrue(isinstance(st2.nested, Struct))
        self.assertTrue(isinstance(st2.nested.ds, Dataset))
        # also make sure no file was created
        self.assertFalse(os.path.exists(r"riptable/tests/temp/save_st/nested.sds"))

        shutil.rmtree(r"riptable/tests/temp/save_st")

    def test_sds_flatten(self):
        ds = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        st = Struct({"ds1": ds, "ds2": ds, "ds3": ds, "arr": arange(5)})
        shutil.rmtree(r"riptable/tests/temp")
        os.makedirs(r"riptable/tests/temp")
        st.save(r"riptable/tests/temp/root_temp")
        st.save(r"riptable/tests/temp/root_temp/st1")

        # garbage directory in nest
        st.save(r"riptable/tests/temp/root_temp/st2")

        try:
            os.makedirs(r"riptable/tests/temp/root_temp/st2/garbagedir")
        except FileExistsError:
            pass

        # garbage file in nest
        st.save(r"riptable/tests/temp/root_temp/st3")
        ds.save(r"riptable/tests/temp/root_temp/st3/garbagefile.sds")

        try:
            os.rename(
                r"riptable/tests/temp/root_temp/st3/garbagefile.sds", r"riptable/tests/temp/root_temp/st3/garbagefile"
            )
        except FileExistsError:
            pass

        sds_flatten(r"riptable/tests/temp/root_temp")
        self.assertFalse(os.path.isdir(r"riptable/tests/temp/root_temp/st1"))
        self.assertTrue(os.path.isfile(r"riptable/tests/temp/root_temp/st1.sds"))
        self.assertTrue(os.path.isfile(r"riptable/tests/temp/root_temp/st1!ds1.sds"))
        self.assertTrue(os.path.isfile(r"riptable/tests/temp/root_temp/st1!ds2.sds"))
        self.assertTrue(os.path.isfile(r"riptable/tests/temp/root_temp/st1!ds3.sds"))
        self.assertTrue(os.path.isdir(r"riptable/tests/temp/root_temp/st2"))
        self.assertTrue(os.path.isdir(r"riptable/tests/temp/root_temp/st3"))

        new_st = load_sds(r"riptable/tests/temp/root_temp", include_all_sds=True)

        s = Struct()
        s["a"] = Dataset({"a": [1, 2, 3]})
        s["b"] = Struct()
        s.save(r"riptable/tests/temp/st4.sds", onefile=True)
        s4 = load_sds([r"riptable/tests/temp/st4.sds", r"riptable/tests/temp/st4.sds"], stack=True)
        self.assertTrue(np.all(s4.a.a == [1, 2, 3, 1, 2, 3]))

        shutil.rmtree(r"riptable/tests/temp/root_temp")

        # ADD TESTS FOR:
        # conflict with temp root
        # conflict with existing file

    # def test_load_h5(self):
    #    try:
    #        import hdf5
    #    except:
    #        print('Could not import hdf5 module. Skipping load_h5 test in test_saveload.py')
    #    else:
    #        # load everything
    #        st = load_h5(r'riptable/tests/test_files/h5data.h5', name='/')
    #        self.assertTrue(isinstance(st, Struct))
    #        self.assertEqual(len(st), 1)

    #        # load 1 struct
    #        xtra = load_h5(r'riptable/tests/test_files/h5data.h5', name='xtra')
    #        self.assertTrue(isinstance(xtra, Struct))
    #        self.assertTrue(len(xtra), 4)

    #        # roundtrip h5 -> riptable -> sds -> riptable
    #        xtra.save(r'riptable/tests/temp/testsaveh5')
    #        xtra2 = Struct.load(r'riptable/tests/temp/testsaveh5')
    #        for k,v in xtra.items():
    #            self.assertTrue(bool(np.all(v == xtra2[k])))
    #        shutil.rmtree(r'riptable/tests/temp/testsaveh5')

    #        # drop short columns
    #        with self.assertWarns(UserWarning):
    #            xtra_ds = load_h5(r'riptable/tests/test_files/h5data.h5', name='xtra', drop_short=True)
    #            self.assertTrue(isinstance(xtra_ds, Dataset))
    #            self.assertTrue(len(xtra_ds), 1)

    def test_all_types(self):

        test_len = 20_000

        ds = Dataset(
            {
                "int8": np.random.randint(low=-128, high=127, size=(test_len), dtype=np.int8),
                "uint8": np.random.randint(low=0, high=255, size=(test_len), dtype=np.uint8),
                "int16": np.random.randint(low=-32768, high=32767, size=(test_len), dtype=np.int16),
                "uint16": np.random.randint(low=0, high=65535, size=(test_len), dtype=np.uint16),
                "int32": np.random.randint(low=-2147483648, high=2147483647, size=(test_len), dtype=np.int32),
                "uint32": np.random.randint(low=0, high=4294967295, size=(test_len), dtype=np.uint32),
                "int64": np.random.randint(
                    low=-9223372036854775808, high=9223372036854775807, size=(test_len), dtype=np.int64
                ),
                "uint64": np.random.randint(low=0, high=18446744073709551615, size=(test_len), dtype=np.uint64),
                "float32": np.random.rand(test_len).astype(np.float32) * 10e9,
                "float64": np.random.rand(test_len).astype(np.float64) * 10e-9,
                "unicode_str": FastArray(np.random.choice(["a", "b", "c"], test_len), unicode=True),
                "bytes_str": np.random.choice([b"a", b"b", b"c"], test_len),
            }
        )

        path = r"riptable/tests/temp/alltypes.sds"

        ds.save(path)
        ds2 = load_sds(path)
        self.assertTrue(list(ds) == list(ds2))
        for k, col in ds2.items():
            self.assertTrue(col.dtype == ds[k].dtype)
            self.assertTrue(bool(np.all(col == ds[k])))

        os.remove(path)

        ds.save(path, compress=False)
        ds2 = load_sds(path)
        self.assertTrue(list(ds) == list(ds2))
        for k, col in ds2.items():
            self.assertTrue(col.dtype == ds[k].dtype)
            self.assertTrue(bool(np.all(col == ds[k])))

        os.remove(path)

    def test_type_flip(self):
        # unless they can't be converted to a non-object single-item numpy array, all python
        # scalars will flip to numpy scalars.
        # any objects will be stored as their string representation.
        st = Struct(
            {
                "empty_unicode": "",
                "empty_bytes": b"",
                "empty_unicode_arr": np.array([""]),
                "empty_bytes_arr": np.array([b""]),
                "nan": np.nan,
                "empty_numeric_array": arange(0),
                "npint8": np.int8(0),
                "npint8": np.uint8(0),
                "npint16": np.int16(0),
                "npuint16": np.uint16(0),
                "npint32": np.int32(0),
                "npuint32": np.uint32(0),
                "npint64": np.int64(0),
                "npuint64": np.uint64(0),
                "npfloat32": np.float32(0),
                "npfloat64": np.float64(0),
                "pyint": 0xFFFFFFFFFFFFFFFFF,  # stored as string, warns
                "pyflt": 3.14,
                "bool": True,
                "npbool": np.bool_(True),
                "npstr_scalar": np.str_("test"),
                "npbytes_scalar": np.bytes_("test"),
                "none": None,
            }
        )
        path = r"riptable/tests/temp/alltypes"
        with self.assertWarns(UserWarning):
            save_sds(path, st)
        st2 = load_sds(path)

        s = st2.empty_unicode
        self.assertEqual(s.__class__.__name__, "str_")

        s = st2.empty_bytes
        self.assertEqual(s.__class__.__name__, "bytes_")

        s = st2.pyflt
        self.assertEqual(s.__class__, np.float64)

        s = st2.bool
        self.assertEqual(s.__class__, np.bool_)

        s = st2.pyint
        self.assertEqual(s, b"295147905179352825855")

        os.remove(r"riptable/tests/temp/alltypes.sds")

    # leave this test here until pdataset has its own save / load
    def test_pdataset_save_warning(self):
        ds = Dataset({"col_" + str(i): np.random.rand(5) for i in range(7)})
        pds = PDataset([ds, ds])
        path = r"riptable/tests/temp/pds.sds"
        with self.assertWarns(UserWarning):
            pds.save(path)

        pds2 = load_sds(path)
        self.assertTrue(pds2.__class__ == Dataset)

        os.remove(path)

    def test_stack_cat_invalid_index(self):
        c = Categorical(["a", "a", "b", "c", "a"])
        c._fa[0] = -128
        ds = Dataset({"catcol": c})
        path = r"riptable/tests/temp/ds1.sds"
        ds.save(path)

        ds2 = load_sds([path, path], stack=True)

        self.assertTrue(isinstance(ds2.catcol, Categorical))
        self.assertEqual(ds2.catcol._fa[0], 0)

        os.remove(path)

    def test_load_multikey_base_0(self):
        arr = np.random.choice(["a", "b", "c"], 10)
        c = Categorical([arr, arr], base_index=0)
        self.assertTrue(isinstance(c, Categorical))
        self.assertTrue(c.ismultikey)
        self.assertEqual(c.base_index, 0)

        path = r"riptable/tests/temp/ds1.sds"
        ds1 = Dataset({"catcol": c})
        ds1.save(path)

        ds2 = load_sds(path)
        c2 = ds2.catcol
        self.assertTrue(isinstance(c2, Categorical))
        self.assertTrue(c2.ismultikey)
        self.assertEqual(c2.base_index, 0)

        self.assertTrue(ds1.catcol.count().equals(ds2.catcol.count()))

        os.remove(path)

    def test_onefile(self):
        ds = Dataset({"arr1": arange(5), "arr2": arange(5)})
        st = Struct({"ds": ds, "arr3": arange(5)})

        folderpath = r"riptable/tests/temp/stfolder"
        st.save(folderpath, onefile=False)
        self.assertTrue(os.path.isdir(folderpath))
        shutil.rmtree(folderpath)

        singlepath = r"riptable/tests/temp/stfile.sds"
        st.save(singlepath, onefile=True)
        self.assertTrue(os.path.isfile(singlepath))
        os.remove(singlepath)

    def test_onefile_folder(self):
        ds = Dataset({"arr1": arange(5), "arr2": arange(5)})
        st = Struct({"ds1": ds, "arr3": arange(5)})

        singlepath = r"riptable/tests/temp/stfile.sds"
        st.save(singlepath, onefile=True)
        ds1 = load_sds(singlepath, folders=["ds1/"])
        self.assertTrue(isinstance(ds1, Dataset))
        self.assertTrue(ds1.equals(ds))

        os.remove(singlepath)

    # these tests are muted until root rebuild is committed
    def _test_rebuild_root(self):
        ds = Dataset({"arr1": arange(5), "arr2": arange(5)})
        st = Struct({"ds1": ds, "arr3": arange(5)})

        folderpath = r"riptable/tests/temp/stfolder"
        st.save(folderpath)
        ds.save(os.path.join(folderpath, "ds2"))

        st2 = load_sds(folderpath)

        self.assertTrue("ds2" in st2)
        self.assertTrue(isinstance(st2.ds2, Dataset))

        shutil.rmtree(folderpath)

    def _test_rebuild_root_onefile(self):
        ds = Dataset({"arr1": arange(5), "arr2": arange(5)})
        st = Struct({"ds1": ds, "arr3": arange(5)})

        folderpath = r"riptable/tests/temp/stfolder"
        st.save(folderpath)
        st.save(os.path.join(folderpath, "st2"), onefile=True)

        st2 = load_sds(folderpath)

        self.assertTrue("st2" in st2)
        self.assertTrue(isinstance(st2.st2, Struct))

        shutil.rmtree(folderpath)

    def _test_rebuild_root_name(self):
        ds = Dataset({"arr1": arange(5), "arr2": arange(5)})
        st = Struct({"ds1": ds, "arr3": arange(5)})

        folderpath = r"riptable/tests/temp/stfolder"
        st.save(folderpath)
        st.save(folderpath, name="st2")

        st2 = load_sds(folderpath)

        self.assertTrue("st2" in st2)
        self.assertTrue(isinstance(st2.st2, Struct))

        shutil.rmtree(folderpath)

    def test_load_filter(self):
        ds = Dataset({_k: list(range(_i * 10, (_i + 1) * 10)) for _i, _k in enumerate(["a", "b", "c", "d", "e"])})
        singlepath = r"riptable/tests/temp/stfile.sds"
        ds.save(singlepath)
        ds1 = load_sds(singlepath, filter=[False, False, True, True, True, False, False, False, False, False])
        ds2 = load_sds(singlepath, filter=arange(2, 5))
        self.assertTrue(np.all(ds1.crc.imatrix_make() == ds2.crc.imatrix_make()))
        os.remove(singlepath)

    # def test_onefile_stack(self):
    #    ds = Dataset({'arr1':arange(5), 'arr2':arange(5)})
    #    st = Struct({'ds1':ds, 'arr3':arange(5)})

    #    paths = [ r'riptable/tests/temp/stfile'+str(i)+'.sds' for i in range(3) ]
    #    for p in paths:
    #        st.save(p, onefile=True)

    #    # disabled, names don't match after stack

    #    stacked = load_sds(paths, folders=['ds1/'], stack=True)
    #    # are we popping this item automatically?
    #    stacked = stacked.ds1
    #    self.assertTrue(isinstance(stacked, PDataset))

    #    pds = PDataset([ds for _ in range(len(paths))])
    #    self.assertTrue(pds.equals(stacked))

    #    for p in paths:
    #        os.remove(p)


# TODO fold test_sds_stack_with_categorical into the more general test_sds_stack
# We will still want to test across various container types, but rt_test_data module should be responsible for that detail
@pytest.mark.parametrize("container_type", [Dataset, Struct])
@pytest.mark.parametrize("stack", [True, False])
@pytest.mark.parametrize("stack_count", [1])  # TODO consider stack_count 0 and 2
@pytest.mark.parametrize("data", [get_all_categorical_data()])
def test_sds_stack_with_categorical(container_type, data, stack, stack_count, tmpdir):
    fn = "test_sds_stack_with_categorical"

    # one SDS file per each container per each data element
    for i, exp in enumerate(data):
        key = f"{container_type.__name__}_{name(exp)}_{i}"
        filename = str(tmpdir.join(key))

        expected = container_type({key: exp})
        save_sds(filename, expected)

        # TODO stack across SDS files that are not identical
        # TODO add case where data is sliced to same length and stored in one container
        actual: Optional[PDataset] = None
        if stack:
            actual = load_sds([filename] * stack_count, stack=stack)
        else:
            actual = load_sds(filename, stack=stack)

        act = actual[key]
        if isinstance(exp, Categorical):
            assert_categorical_equal(act, exp, verbose=True)
        elif isinstance(exp, Struct):
            assert act.equals(exp)
        else:
            pytest.fail(
                f"{fn}: assertions not implemented for data type {type(data)}\n"
                + f"expected of type {type(exp)}\n{repr(exp)}"
                + f"actual of type {type(act)}\n{repr(act)}"
            )


@pytest.mark.parametrize(
    "stack",
    [
        pytest.param(
            True,
            marks=pytest.mark.xfail(
                reason="RIP-483 - Raises 'ValueError: SDS stacking only implemented for Datasets. Must provide folders list if loading from multiple Struct directories.'"
            ),
        ),
        False,
        pytest.param(
            None,
            marks=pytest.mark.xfail(
                reason="RIP-482 - Saving a Struct that contains a Categorical with stacking set to `None` and `stack_count=1` modifies the publically accessible category_dict by modifying the key from `CAT` to `categorical_bytes`. To reproduce, remove this `pytest.mark.xfail`."
            ),
        ),
    ],
)
@pytest.mark.parametrize(
    "stack_count",
    [
        1,
        pytest.param(0, marks=pytest.mark.xfail(reason="Same as reason for param `2` below")),
        pytest.param(
            2,
            marks=pytest.mark.xfail(
                reason="RIP-482 - Saving a Struct that contains a Categorical with stacking set to `False` and `stack_count=2` modifies the publically accessible category_dict by modifying the key from `CAT` to `categorical_bytes`. To reproduce, remove this `pytest.mark.xfail`."
            ),
        ),
    ],
)
@pytest.mark.parametrize("data", [load_test_data()])
def test_sds_stack(data, stack, stack_count, tmpdir):
    fn = "test_sds_stack"

    # one SDS file per each container per each data element
    key = f"{fn}_{id(data)}"
    filename = str(tmpdir.join(key))

    save_sds(filename, data)

    # TODO stack across SDS files that are not identical
    # TODO add case where data is sliced to same length and stored in one container
    actual: Optional[PDataset] = None
    if stack:
        actual = load_sds([filename] * stack_count, stack=stack)
    else:
        actual = load_sds(filename, stack=stack)

    for k, v in data.items():
        exp, act = data[k], actual[k]
        err_msg = (
            f"{fn}: failed equality check for {k}\n"
            + f"expected of type {type(data)}\n{repr(data)}"
            + f"actual of type {type(act)}\n{repr(act)}"
        )
        if isinstance(exp, Categorical):
            assert_categorical_equal(act, exp, verbose=True)
        elif isinstance(exp, FastArray):
            assert_array_equal_(act, exp, err_msg=err_msg)
        elif isinstance(data, Struct):
            assert act.equals(exp), err_msg
        else:
            pytest.fail(f"{fn}: assertions not implemented for data type {type(exp)}\n" + err_msg)


@pytest.mark.parametrize(
    "data",
    [
        Struct(),
        Struct(Struct()),
        Struct(Multiset()),
        Struct({"s": Struct()}),
        Struct({"d": Dataset({})}),
        pytest.param(
            Struct({"m": Multiset()}),
            marks=pytest.mark.xfail(reason="_sds_from_tree: 'NoneType' object cannot be interpreted as an integer"),
        ),
        pytest.param(
            Multiset({"d": Dataset()}),
            marks=pytest.mark.xfail(reason="_sds_from_tree: 'NoneType' object cannot be interpreted as an integer"),
        ),
    ],
)
def test_empty_roundtrip(data, tmpdir):
    dir = tmpdir.mkdir("test_empty_roundtrip")
    p = str(dir.join(f"tmp_{id(data)}.sds"))
    save_sds(p, data)
    data2 = load_sds(p)
    assert isinstance(data2, data.__class__)
    assert data.shape == data2.shape, f"saved shape should equal original shape"


@pytest.mark.parametrize("rng_seed", [590518374])
@pytest.mark.parametrize(
    "include_date",
    [
        pytest.param(
            False,
            marks=pytest.mark.xfail(
                reason="Test fails due to the issue described in https://github.com/rtosholdings/riptable/issues/138"
            ),
        ),
        pytest.param(
            True,
            marks=pytest.mark.xfail(
                reason="Categorical with Date-typed category does not currently preserve the type (Date) of the category data during round-tripping through SDS."
            ),
        ),
    ],
)
@pytest.mark.parametrize("explicit_col_load_list", [False, True])
def test_multi_section_sds_stacked_load_of_single_cat(
    tmp_path, request, rng_seed: int, include_date: bool, explicit_col_load_list: bool
) -> None:
    """
    Test that loading a column(s) from a single SDS file containing multiple 'sections'
    (due to appending or using rt.sds_concat()) works as expected.
    The column(s) should have the correct length and be gap-filled wherever the column(s) do(es)
    not appear in some section(s) of the file.

    This test is a repro/test for https://github.com/rtosholdings/riptable/issues/138
    """
    rng = default_rng(rng_seed)

    # Randomly generate the rowcounts for the test datasets we'll generate.
    ds_count = 10
    min_rowcount = 5
    max_rowcount = 20
    ds_rowcounts = rng.integers(low=min_rowcount, high=max_rowcount, size=ds_count, endpoint=True)

    # Generate the test datasets.
    test_dss = [
        create_test_dataset(rng=rng, rowcount=int(x), include_dict_cat=True, include_date=include_date)
        for x in ds_rowcounts
    ]
    test_ds_keys = set(test_dss[0].keys())

    # For some of the test datasets, drop one or more columns --
    # this lets us test SDS performs gap-filling as expected when loading.
    drop_cat_col_prob = 0.3
    fa_col_name = "open_price"
    cat_col_name = "deliverable_symbol"
    dropped_col_names = [fa_col_name, cat_col_name]
    for dropped_col_name in dropped_col_names:
        assert (
            dropped_col_name in test_dss[0]
        )  # sanity check to make sure we don't attempt to drop a non-existent column.
    drop_cat_col = rng.choice(
        2, size=len(test_dss), replace=True, p=[drop_cat_col_prob, 1.0 - drop_cat_col_prob]
    ).astype(np.bool_)
    assert 0 < drop_cat_col.sum() < len(test_dss), "Bad random data for drop-column flags."

    for do_drop_cat, test_ds in zip(drop_cat_col, test_dss):
        if do_drop_cat:
            test_ds.col_remove(dropped_col_names)

    # Save the datasets to a single SDS file by creating the file then appending to it.
    data_path = tmp_path / f"{request.node.originalname}-{rng_seed}.sds"
    test_dss[0].save(data_path, overwrite=True)
    for i in range(1, len(test_dss)):
        test_dss[i].save(data_path, append=f"test_section_{i}")

    # Row-concat the in-memory Datasets so we have an expected result
    # to compare the output of rt.load_sds(..., stack=True) against.
    inmem_stacked_ds = rt.Dataset.concat_rows(test_dss)
    assert test_ds_keys == set(inmem_stacked_ds.keys())

    # First, check whether the dropped column(s) is gap-filled as expected when we load the entire Dataset.
    include_keys = inmem_stacked_ds.keys() if explicit_col_load_list else None
    sds_stacked_allcols_ds = rt.load_sds(data_path, include=include_keys, stack=True)
    assert test_ds_keys == set(sds_stacked_allcols_ds.keys())
    assert inmem_stacked_ds.get_nrows() == sds_stacked_allcols_ds.get_nrows()
    for col_name in test_ds_keys:
        inmem_col = inmem_stacked_ds[col_name]
        sds_loaded_col = sds_stacked_allcols_ds[col_name]

        # Verify the columns match up.
        assert_array_or_cat_equal(
            inmem_col,
            sds_loaded_col,
            err_msg=f"The array data for column '{col_name}' does not match the expected result.",
            # Categorical equality comparison options.
            relaxed_cat_check=True,
        )

    # For debugging purposes, determine how many rows are in the datasets where the column was dropped.
    dropped_row_count = sum(map(lambda do_drop, ds: ds.get_nrows() if do_drop else 0, drop_cat_col, test_dss))
    print(f"Datasets where the target column was dropped account for {dropped_row_count} rows.")

    # Try to load the column(s) we randomly dropped from some of the Datasets.
    dropped_col_name_lists = []
    # First try loading all of the dropped columns plus one column present in all sections.
    dropped_col_name_lists.append(["close_price", *dropped_col_names])
    dropped_col_name_lists.append(dropped_col_names)
    dropped_col_name_lists.extend([[x] for x in dropped_col_names])
    for dropped_col_name_list in dropped_col_name_lists:
        with rt.load_sds(data_path, include=dropped_col_name_list, stack=True) as sds_stacked_dropcol_ds:
            assert sds_stacked_dropcol_ds.get_ncols() == len(dropped_col_name_list)
            for dropped_col_name in dropped_col_name_list:
                assert dropped_col_name in sds_stacked_dropcol_ds
            for dropped_col_name in dropped_col_name_list:
                assert_array_or_cat_equal(
                    inmem_stacked_ds[dropped_col_name],
                    sds_stacked_dropcol_ds[dropped_col_name],
                    err_msg=f"The array data for column '{dropped_col_name}' does not match the expected result.\tDropped columns: {dropped_col_name_list}",
                    # Categorical equality comparison options.
                    relaxed_cat_check=True,
                )


@pytest.mark.parametrize("rng_seed", [558902840])
@pytest.mark.parametrize(
    "include_date",
    [
        pytest.param(
            False,
            marks=pytest.mark.xfail(
                reason="Test fails due to the issue described in https://github.com/rtosholdings/riptable/issues/163"
            ),
        )
    ],
)
@pytest.mark.parametrize("explicit_col_load_list", [False, True])
@pytest.mark.parametrize("randomize_start_mask", [False, True])
@pytest.mark.parametrize(
    "mask_dropped_section_prop",
    [
        pytest.param(None, id="No dropped sections completely masked out"),
        pytest.param(0.01, id="One dropped section completely masked out."),
        pytest.param(1.0, id="All dropped sections completely masked out."),
    ],
)
@pytest.mark.parametrize(
    "mask_undropped_section_prop",
    [
        pytest.param(None, id="No undropped sections completely masked out"),
        pytest.param(0.01, id="One undropped section completely masked out."),
        pytest.param(1.0, id="All undropped sections completely masked out."),
    ],
)
def test_multi_section_sds_stacked_load_single_array_filtered(
    tmp_path,
    request,
    rng_seed: int,
    include_date: bool,
    explicit_col_load_list: bool,
    randomize_start_mask: bool,
    mask_dropped_section_prop: Optional[float],
    mask_undropped_section_prop: Optional[float],
) -> None:
    """
    Test that loading a column(s) from a single SDS file containing multiple 'sections'
    (due to appending or using rt.sds_concat()) works as expected when a mask/filter is provided
    to `rt.load_sds()`.
    The column(s) should have the correct length and be gap-filled wherever the column(s) do(es)
    not appear in some section(s) of the file.

    This test is a repro/test for https://github.com/rtosholdings/riptable/issues/163
    """
    rng = default_rng(rng_seed)

    # Randomly generate the rowcounts for the test datasets we'll generate.
    ds_count = 10
    min_rowcount = 5
    max_rowcount = 20
    ds_rowcounts = rng.integers(low=min_rowcount, high=max_rowcount, size=ds_count, endpoint=True)

    # Generate the test datasets.
    test_dss = [
        create_test_dataset(rng=rng, rowcount=int(x), include_dict_cat=True, include_date=include_date)
        for x in ds_rowcounts
    ]
    test_ds_keys = set(test_dss[0].keys())

    # For some of the test datasets, drop one or more columns --
    # this lets us test SDS performs gap-filling as expected when loading.
    drop_cat_col_prob = 0.3
    fa_col_name = "open_price"
    cat_col_name = "deliverable_symbol"
    dropped_col_names = [fa_col_name, cat_col_name]
    for dropped_col_name in dropped_col_names:
        assert (
            dropped_col_name in test_dss[0]
        )  # sanity check to make sure we don't attempt to drop a non-existent column.
    drop_cat_col = rng.choice(
        2, size=len(test_dss), replace=True, p=[drop_cat_col_prob, 1.0 - drop_cat_col_prob]
    ).astype(np.bool_)
    assert 0 < drop_cat_col.sum() < len(test_dss), "Bad random data for drop-column flags."

    for do_drop_cat, test_ds in zip(drop_cat_col, test_dss):
        if do_drop_cat:
            test_ds.col_remove(dropped_col_names)

    # Save the datasets to a single SDS file by creating the file then appending to it.
    data_path = tmp_path / f"{request.node.originalname}-{rng_seed}.sds"
    test_dss[0].save(data_path, overwrite=True)
    for i in range(1, len(test_dss)):
        test_dss[i].save(data_path, append=f"test_section_{i}")

    # Create the mask we'll use to filter the data -- both in-memory and when we call rt.load_sds().
    fully_masked_section_flags = rt.ones(len(test_dss), dtype=np.bool_)
    if mask_dropped_section_prop is not None:
        dropped_section_idxs = np.where(drop_cat_col)
        maskout_dropped_section_count = int(np.ceil(mask_dropped_section_prop * len(dropped_section_idxs)))
        maskout_dropped_section_idxs = rng.choice(dropped_section_idxs, size=maskout_dropped_section_count)
        fully_masked_section_flags[maskout_dropped_section_idxs] = False

    if mask_undropped_section_prop is not None:
        undropped_section_idxs = np.where(np.logical_not(drop_cat_col))
        maskout_undropped_section_count = int(np.ceil(mask_undropped_section_prop * len(undropped_section_idxs)))
        maskout_undropped_section_idxs = rng.choice(undropped_section_idxs, size=maskout_undropped_section_count)
        fully_masked_section_flags[maskout_undropped_section_idxs] = False

    data_mask = np.repeat(fully_masked_section_flags, ds_rowcounts)
    if randomize_start_mask:
        total_rowcount = sum(ds.get_nrows() for ds in test_dss)
        row_maskout_prob = 0.2  # probability for each row that it's masked OUT
        overlay_mask = rng.choice(
            2, size=total_rowcount, replace=True, p=[row_maskout_prob, 1.0 - row_maskout_prob]
        ).astype(np.bool_)
        rt.mask_andi(data_mask, overlay_mask)

    # Row-concat the in-memory Datasets so we have an expected result
    # to compare the output of rt.load_sds(..., stack=True) against.
    inmem_stacked_ds = rt.Dataset.concat_rows(test_dss)
    inmem_stacked_ds = inmem_stacked_ds[data_mask, :]
    assert test_ds_keys == set(inmem_stacked_ds.keys())

    # First, check whether the dropped column(s) is gap-filled as expected when we load the entire Dataset.
    include_keys = inmem_stacked_ds.keys() if explicit_col_load_list else None
    sds_stacked_allcols_ds = rt.load_sds(data_path, include=include_keys, stack=True, filter=data_mask)
    assert test_ds_keys == set(sds_stacked_allcols_ds.keys())
    assert inmem_stacked_ds.get_nrows() == sds_stacked_allcols_ds.get_nrows()
    for col_name in test_ds_keys:
        inmem_col = inmem_stacked_ds[col_name]
        sds_loaded_col = sds_stacked_allcols_ds[col_name]

        # Verify the columns match up.
        assert_array_or_cat_equal(
            inmem_col,
            sds_loaded_col,
            err_msg=f"The array data for column '{col_name}' does not match the expected result.",
            # Categorical equality comparison options.
            relaxed_cat_check=True,
            # Don't bother checking category names until SDS round-trips them.
            check_cat_names=False,
            # TEMP: assert_array_or_cat_equal() still checks names right now, unless check_cat_types is disabled.
            check_cat_types=False,
        )

    # For debugging purposes, determine how many rows are in the datasets where the column was dropped.
    dropped_row_count = sum(map(lambda do_drop, ds: ds.get_nrows() if do_drop else 0, drop_cat_col, test_dss))
    print(
        f"Datasets where the target column was dropped account for {dropped_row_count}/{ds_rowcounts.sum()} rows. (Mask selected {data_mask.sum()}/{len(data_mask)} rows.)"
    )

    # Try to load the column(s) we randomly dropped from some of the Datasets.
    dropped_col_name_lists = []
    # First try loading all of the dropped columns plus one column present in all sections.
    dropped_col_name_lists.append(["close_price", *dropped_col_names])
    dropped_col_name_lists.append(dropped_col_names)
    dropped_col_name_lists.extend([[x] for x in dropped_col_names])
    for dropped_col_name_list in dropped_col_name_lists:
        with rt.load_sds(
            data_path, include=dropped_col_name_list, stack=True, filter=data_mask
        ) as sds_stacked_dropcol_ds:
            assert sds_stacked_dropcol_ds.get_ncols() == len(dropped_col_name_list)
            for dropped_col_name in dropped_col_name_list:
                assert dropped_col_name in sds_stacked_dropcol_ds
            for dropped_col_name in dropped_col_name_list:
                assert_array_or_cat_equal(
                    inmem_stacked_ds[dropped_col_name],
                    sds_stacked_dropcol_ds[dropped_col_name],
                    err_msg=f"The array data for column '{dropped_col_name}' does not match the expected result.\tDropped columns: {dropped_col_name_list}",
                    # Categorical equality comparison options.
                    relaxed_cat_check=True,
                    # Don't bother checking category names until SDS round-trips them.
                    check_cat_names=False,
                    # TEMP: assert_array_or_cat_equal() still checks names right now, unless check_cat_types is disabled.
                    check_cat_types=False,
                )


if __name__ == "__main__":
    tester = unittest.main()



from enum import IntEnum

import pytest

from riptable import *
from riptable.tests.utils import LikertDecision


class TestCategoricalChangeCategories:
    def test_add_mapping(self):
        codes = [44, 133, 75, 1, 144, 144, 1, 44, 75]
        c = Categorical(codes, LikertDecision)
        c.mapping_add(50, "Fifty")
        new_mapping = c.category_mapping
        result = new_mapping.get(50, False)
        assert result == "Fifty"

        with pytest.raises(ValueError):
            c.mapping_add(1, "One")

        with pytest.raises(TypeError):
            c.mapping_add("notanint", "Two")

        codes = [33, 133, 75, 1, 144, 144, 1, 33, 75]
        c = Categorical(codes, LikertDecision)
        assert c[0] == "!<33>"
        c.mapping_add(33, "ThirtyThree")
        assert c[0] == "ThirtyThree"

    def test_remove_mapping(self):
        codes = [44, 133, 75, 1, 144, 144, 1, 44, 75]
        c = Categorical(codes, LikertDecision)
        c.mapping_remove(44)
        assert c[0] == "!<44>"
        assert c[7] == "!<44>"
        new_mapping = c.category_mapping
        removed = new_mapping.get(44, True)
        assert removed

        with pytest.raises(ValueError):
            c.mapping_remove(33)

    def test_replace_mapping(self):
        codes = [44, 133, 75, 1, 144, 144, 1, 44, 75]
        c = Categorical(codes, LikertDecision)
        c.mapping_replace(1, "One")
        assert c[3] == "One"
        new_mapping = c.category_mapping
        assert new_mapping[1] == "One"

        with pytest.raises(ValueError):
            c.mapping_replace(50, "Fifty")

    def test_new_mapping(self):
        class MappingNew(IntEnum):
            FourFour = 44
            OneThreeThree = 133
            SevenFive = 75
            ThreeSix = 36
            OneFourFour = 144

        codes = [44, 133, 75, 1, 144, 144, 1, 44, 75]
        c = Categorical(codes, LikertDecision)
        assert c[3] == "StronglyDisagree"
        c.mapping_new(MappingNew)
        assert c[0] == "FourFour"
        assert c[3] == "!<1>"

        with pytest.raises(TypeError):
            c.mapping_new(np.arange(10))

    # TODO: add more tests for different types of stringlist categoricals
    def test_category_add(self):
        c = Categorical(["b", "b", "c", "d", "e", "b", "c"])
        orig_fa = c._fa.copy()
        c.category_add("a")
        first_match = bool(np.all(orig_fa == c._fa))
        assert first_match
        new_cat_array = c.category_array
        assert len(new_cat_array) == 5
        assert new_cat_array[-1] == b"a"

        with pytest.raises(ValueError):
            c.category_add("a")

    def test_category_remove(self):
        c = Categorical(["b", "b", "c", "d", "e", "b", "c"])
        c.category_remove("b")
        c.filtered_set_name("Inv")
        assert c[1] == "Inv"
        assert c._fa[1] == 0
        assert c._fa[2] == 1
        assert len(c.category_array) == 3

        with pytest.raises(ValueError):
            c.category_remove("b")

    def test_category_replace(self):
        c = Categorical(["b", "b", "c", "d", "e", "b", "c"])
        c.category_replace("b", "a")
        assert c.category_array[0] == b"a"
        assert c[0] == "a"
        assert not (b"b" in c.category_array)
        c.category_replace("a", "d")
        assert len(c.category_array) == 4
        assert not c._ordered

        with pytest.raises(ValueError):
            c.category_replace("z", "a")



import unittest

from riptable import *


def arr_eq(a, b):
    return bool(np.all(a == b))


def cat_eq(c1, c2):
    # for single key
    # TODO: compare other attributes
    inst_eq = arr_eq(c1._fa, c2._fa)
    cats_eq = arr_eq(c1.category_array, c2.category_array)
    return inst_eq and cats_eq


class Flatten_Test(unittest.TestCase):
    # TODO pull nested Struct into testing data for reuse in tooling integration tests
    def test_nested(self):
        st1 = Struct(
            {
                "cat1": Categorical(["a", "b", "c", "d"]),
                "arr1": arange(5),
                "ds1": Dataset({"col_" + str(i): np.random.rand(5) for i in range(5)}),
                "arr2": arange(10),
                "test_string": "this is my string",
                "array3": arange(30).astype(np.int8),
                "nested_struct1": Struct(
                    {
                        "set1": Dataset({"col_" + str(i): np.random.randint(5) for i in range(3)}),
                        "set2": Dataset({"col_" + str(i): np.random.choice(["a", "b", "c"], 10) for i in range(2)}),
                        "nested_2": Struct(
                            {
                                "c1": Categorical(arange(5)),
                                "c2": Categorical(["aaa", "bbbb", "cccc"]),
                                "leaf_dataset": Dataset({"col_" + str(i): arange(5) for i in range(3)}),
                                "array4": arange(20).astype(np.uint64),
                                "dtn1": DateTimeNano.random(20),
                                "nested_3": Dataset(
                                    {
                                        "CAT": Categorical(np.random.choice(["X", "Y", "Z"], 17)),
                                        "DTN": DateTimeNano.random(17),
                                        "DATE": Date(np.random.randint(15000, 20000, 17)),
                                        "TSPAN": TimeSpan(
                                            np.random.randint(
                                                0,
                                                1_000_000_000 * 60 * 60 * 24,
                                                17,
                                                dtype=np.int64,
                                            )
                                        ),
                                        "DSPAN": DateSpan(np.random.randint(0, 365, 17)),
                                    }
                                ),
                            }
                        ),
                        "set3": Dataset({"col_" + str(i): np.random.choice([True, False], 10) for i in range(4)}),
                    }
                ),
                "ds2": Dataset({"heading_" + str(i): np.random.rand(5) for i in range(3)}),
                "int1": 5,
                "float1": 7.0,
                "cat2": Categorical(["a", "b", "c", "d"]),
            }
        )

        st_flattened = st1.flatten()
        st2 = st_flattened.flatten_undo()

        # top level keys are the same, in the same order
        self.assertTrue(arr_eq(list(st1), list(st2)))

        # single datasets of normal arrays
        self.assertTrue(isinstance(st2.ds1, Dataset))
        self.assertTrue(st1.ds1.equals(st2.ds1))
        self.assertTrue(isinstance(st2.ds2, Dataset))
        self.assertTrue(st1.ds2.equals(st2.ds2))

        # top level arrays
        self.assertTrue(isinstance(st2.cat1, Categorical))
        self.assertTrue(cat_eq(st1.cat1, st2.cat1))
        self.assertTrue(arr_eq(st1.arr1, st2.arr1))
        self.assertTrue(arr_eq(st1.arr2, st2.arr2))
        self.assertTrue(isinstance(st2.test_string, str))
        self.assertEqual(st1.test_string, st2.test_string)
        self.assertTrue(arr_eq(st1.array3, st2.array3))
        self.assertTrue(isinstance(st2.cat2, Categorical))
        self.assertTrue(cat_eq(st1.cat2, st2.cat2))

        # top level scalars
        self.assertTrue(isinstance(st2.int1, (int, np.integer)))
        self.assertEqual(st1.int1, st2.int1)
        self.assertTrue(isinstance(st2.float1, (float, np.floating)))
        self.assertEqual(st1.float1, st2.float1)

        # main nested struct
        self.assertTrue(isinstance(st2.nested_struct1, Struct))
        self.assertTrue(arr_eq(list(st1.nested_struct1), list(st2.nested_struct1)))

        # dataset of ints
        self.assertTrue(isinstance(st2.nested_struct1.set1, Dataset))
        self.assertTrue(st1.nested_struct1.set1.equals(st2.nested_struct1.set1))

        # dataset of bytes
        self.assertTrue(isinstance(st2.nested_struct1.set2, Dataset))
        self.assertTrue(st1.nested_struct1.set2.equals(st2.nested_struct1.set2))

        # dataset of bools
        self.assertTrue(isinstance(st2.nested_struct1.set2, Dataset))
        self.assertTrue(st1.nested_struct1.set2.equals(st2.nested_struct1.set2))

        # another nested level
        nest1 = st1.nested_struct1.nested_2
        nest2 = st2.nested_struct1.nested_2
        self.assertTrue(isinstance(nest2, Struct))
        self.assertTrue(arr_eq(list(nest1), list(nest2)))

        # arrays
        self.assertTrue(cat_eq(nest1.c1, nest2.c1))
        self.assertTrue(cat_eq(nest1.c2, nest2.c2))
        self.assertTrue(arr_eq(nest1.array4, nest2.array4))
        self.assertTrue(isinstance(nest2.dtn1, DateTimeNano))
        self.assertTrue(arr_eq(nest1.dtn1, nest2.dtn1))

        # datasets
        self.assertTrue(isinstance(nest2.leaf_dataset, Dataset))
        self.assertTrue(nest1.leaf_dataset.equals(nest2.leaf_dataset))

        # final nesting, all special types
        DS1 = nest1.nested_3
        DS2 = nest2.nested_3
        self.assertTrue(isinstance(DS2, Dataset))
        self.assertTrue(arr_eq(list(DS1), list(DS2)))

        self.assertTrue(cat_eq(DS1.CAT, DS2.CAT))
        self.assertTrue(isinstance(DS2.DTN, DateTimeNano))
        self.assertTrue(arr_eq(DS1.DTN, DS2.DTN))

        # TODO: compare tz info, etc.
        self.assertTrue(isinstance(DS2.DATE, Date))
        self.assertTrue(arr_eq(DS1.DATE, DS2.DATE))
        self.assertTrue(isinstance(DS2.TSPAN, TimeSpan))
        self.assertTrue(arr_eq(DS1.TSPAN, DS2.TSPAN))
        self.assertTrue(isinstance(DS2.DSPAN, DateSpan))
        self.assertTrue(arr_eq(DS1.DSPAN, DS2.DSPAN))

    # def test_separator(self):
    #    ds = Dataset({'colA':arange(3), 'colB':arange(10,13)})
    #    st = Struct({'arr':arange(10), 'ds1':ds})
    #    flat = st.flatten()
    #    flatcols = ['arr', 'ds1/', 'ds1/colA', 'ds1/colB']
    #    self.assertTrue(arr_eq(list(flat), flatcols))


if __name__ == "__main__":
    tester = unittest.main()
